# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
#

import ftplib
import os
import urllib
from urllib.request import Request

import requests

from virtinst import log

_URLPREFIX = "https://virtinst-testsuite.example/"
_MOCK_TOPDIR = os.path.dirname(__file__) + "/data/urldetect/"


def make_mock_input_url(distro):
    return _URLPREFIX + distro


def _map_mock_url_to_file(url):
    if url.startswith(_URLPREFIX):
        path = url[len(_URLPREFIX) :]
        fn = _MOCK_TOPDIR + path
        if not os.path.exists(os.path.dirname(fn)):
            raise RuntimeError("Expected mock file not found: %s" % fn)

    elif url.endswith("treeinfo"):
        # If the url is requesting treeinfo, give a fake treeinfo from
        # our testsuite data
        fn = "%s/data/fakemedia/fakerhel6tree/.treeinfo" % os.path.abspath(
            os.path.dirname(__file__)
        )
    else:
        # Otherwise just copy this file
        fn = os.path.abspath(__file__)

    return os.path.abspath(fn)


class _MockRequestsResponse:
    def __init__(self, url):
        log.debug("mocking requests session for url=%s", url)
        fn = _map_mock_url_to_file(url)
        self._content = open(fn).read()
        self.headers = {"content-length": len(self._content)}

    def raise_for_status(self):
        pass

    def iter_content(self, *args, **kwargs):
        dummy = args
        dummy = kwargs
        return [self._content.encode("utf-8")]


class _MockRequestsSession:
    def close(self):
        pass

    def head(self, url, *args, **kwargs):
        dummy = args
        dummy = kwargs
        return _MockRequestsResponse(url)

    def get(self, url, *args, **kwargs):
        dummy = args
        dummy = kwargs
        if "testsuitefail" in url:
            raise RuntimeError("testsuitefail seen, raising mock error")
        return _MockRequestsResponse(url)


class _MockFTPSession:
    def connect(self, *args, **kwargs):
        pass

    def login(self, *args, **kwargs):
        pass

    def voidcmd(self, *args, **kwargs):
        pass

    def quit(self, *args, **kwargs):
        pass

    def size(self, url):
        path = _map_mock_url_to_file(url)
        return os.path.getsize(path)


def _MockUrllibRequest(url):
    url = _map_mock_url_to_file(url)
    url = "file://" + url
    return Request(url)


def setup_mock():
    requests.Session = _MockRequestsSession
    ftplib.FTP = _MockFTPSession
    urllib.request.Request = _MockUrllibRequest



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .device import Device
from ..xmlbuilder import XMLProperty


class DeviceAudio(Device):
    XML_NAME = "audio"

    type = XMLProperty("./@type")
    id = XMLProperty("./@id")



#
# Copyright 2008, 2013, 2015 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import threading

import libvirt

from . import generatename
from . import progress
from .logger import log
from .xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty
from . import xmlutil


_DEFAULT_DEV_TARGET = "/dev"
_DEFAULT_SCSI_TARGET = "/dev/disk/by-path"
_DEFAULT_MPATH_TARGET = "/dev/mapper"


class _StoragePermissions(XMLBuilder):
    XML_NAME = "permissions"
    _XML_PROP_ORDER = ["mode", "owner", "group", "label"]

    mode = XMLProperty("./mode")
    owner = XMLProperty("./owner")
    group = XMLProperty("./group")
    label = XMLProperty("./label")


class _StorageObject(XMLBuilder):
    """
    Base class for building any libvirt storage object.

    Meaningless to directly instantiate.
    """

    ##############
    # Properties #
    ##############

    name = XMLProperty("./name")
    permissions = XMLChildProperty(_StoragePermissions, relative_xpath="./target", is_single=True)


def _preferred_default_pool_path(conn):
    root = conn.get_libvirt_data_root_dir()
    return os.path.join(root, "images")


def _lookup_poolxml_by_path(conn, path):
    for poolxml in conn.fetch_all_pools():
        xml_path = poolxml.target_path
        if xml_path is not None and os.path.abspath(xml_path) == path:
            return poolxml
    return None


class _Host(XMLBuilder):
    _XML_PROP_ORDER = ["name", "port"]
    XML_NAME = "host"

    name = XMLProperty("./@name")
    port = XMLProperty("./@port", is_int=True)


class StoragePool(_StorageObject):
    """
    Base class for building and installing libvirt storage pool xml
    """

    TYPE_DIR = "dir"
    TYPE_FS = "fs"
    TYPE_NETFS = "netfs"
    TYPE_LOGICAL = "logical"
    TYPE_DISK = "disk"
    TYPE_ISCSI = "iscsi"
    TYPE_SCSI = "scsi"
    TYPE_MPATH = "mpath"
    TYPE_GLUSTER = "gluster"
    TYPE_RBD = "rbd"
    TYPE_ZFS = "zfs"

    @staticmethod
    def pool_list_from_sources(conn, pool_type):
        """
        Return a list of StoragePool instances built from libvirt's pool
        source enumeration (if supported).

        :param conn: Libvirt connection
        :param pool_type: Pool type string from I{Types}
        """
        source_xml = "<source/>"

        try:
            xml = conn.findStoragePoolSources(pool_type, source_xml, 0)
        except Exception as e:  # pragma: no cover
            if conn.support.is_error_nosupport(e):
                return []
            raise

        log.debug("Libvirt returned pool sources XML:\n%s", xml)

        import xml.etree.ElementTree as ET

        root = ET.fromstring(xml)

        # We implicitly only support this for pool TYPE_LOGICAL
        ret = [e.text for e in root.findall("./source/name")]

        log.debug("Sources returning: %s", ret)
        return ret

    @staticmethod
    def lookup_default_pool(conn):
        """
        Helper to lookup the default pool. It will return one of
        * The pool named 'default'
        * If that doesn't exist, the pool pointing to the default path
        * Otherwise None
        """
        name = "default"
        path = _preferred_default_pool_path(conn)

        poolxml = None
        for trypool in conn.fetch_all_pools():
            if trypool.name == name:
                poolxml = trypool
                break
        else:
            poolxml = _lookup_poolxml_by_path(conn, path)

        if poolxml:
            log.debug("Found default pool name=%s target=%s", poolxml.name, poolxml.target_path)
        return poolxml

    @staticmethod
    def build_default_pool(conn):
        """
        Attempt to lookup the 'default' pool, but if it doesn't exist,
        create it
        """
        poolxml = StoragePool.lookup_default_pool(conn)
        if poolxml:
            return poolxml

        try:
            name = "default"
            path = _preferred_default_pool_path(conn)
            log.debug("Attempting to build default pool with target '%s'", path)
            defpool = StoragePool(conn)
            defpool.type = defpool.TYPE_DIR
            defpool.name = name
            defpool.target_path = path
            defpool.install(build=True, create=True, autostart=True)
            return defpool
        except Exception as e:  # pragma: no cover
            log.debug("Error building default pool", exc_info=True)
            msg = _("Couldn't create default storage pool '%(path)s': %(error)s") % {
                "path": path,
                "error": str(e),
            }
            raise RuntimeError(msg) from None

    @staticmethod
    def lookup_pool_by_path(conn, path):
        """
        Return the first pool with matching matching target path.
        return the first we find, active or inactive. This iterates over
        all pools and dumps their xml, so it is NOT quick.

        :returns: virStoragePool object if found, None otherwise
        """
        poolxml = _lookup_poolxml_by_path(conn, path)
        if not poolxml:
            return None
        return conn.storagePoolLookupByName(poolxml.name)

    @staticmethod
    def find_free_name(conn, basename, **kwargs):
        """
        Finds a name similar (or equal) to passed 'basename' that is not
        in use by another pool. Extra params are passed to generate_name
        """

        def cb(name):
            for pool in conn.fetch_all_pools():
                if pool.name == name:
                    return True
            return False

        return generatename.generate_name(basename, cb, **kwargs)

    @staticmethod
    def ensure_pool_is_running(pool_object, refresh=False):
        """
        If the passed vmmStoragePool isn't running, start it.

        :param pool_object: vmmStoragePool to check/start
        :param refresh: If True, run refresh() as well
        """
        if pool_object.info()[0] != libvirt.VIR_STORAGE_POOL_RUNNING:
            log.debug("starting pool=%s", pool_object.name())
            pool_object.create(0)
        if refresh:
            log.debug("refreshing pool=%s", pool_object.name())
            pool_object.refresh(0)

    ######################
    # Validation helpers #
    ######################

    @staticmethod
    def validate_name(conn, name):
        XMLBuilder.validate_generic_name(_("Storage object"), name)

        try:
            conn.storagePoolLookupByName(name)
        except libvirt.libvirtError:
            return
        raise ValueError(_("Name '%s' already in use by another pool.") % name)  # pragma: no cover

    def default_target_path(self):
        if not self.supports_target_path():
            return None
        if self.type == self.TYPE_DIR or self.type == self.TYPE_NETFS or self.type == self.TYPE_FS:
            return os.path.join(_preferred_default_pool_path(self.conn), self.name)
        if self.type == self.TYPE_ISCSI or self.type == self.TYPE_SCSI:
            return _DEFAULT_SCSI_TARGET
        if self.type == self.TYPE_MPATH:
            return _DEFAULT_MPATH_TARGET

    def _type_to_source_prop(self):
        if self.type == self.TYPE_NETFS or self.type == self.TYPE_GLUSTER:
            return "_source_dir"
        elif self.type == self.TYPE_SCSI:
            return "_source_adapter"
        else:
            return "_source_device"

    def _get_source(self):
        return getattr(self, self._type_to_source_prop())

    def _set_source(self, val):
        return setattr(self, self._type_to_source_prop(), val)

    source_path = property(_get_source, _set_source)

    def default_source_name(self):
        if not self.supports_source_name():
            return None

        if self.type == StoragePool.TYPE_RBD:
            return "rbd"
        if self.type == StoragePool.TYPE_GLUSTER:
            return "gv0"

    ##############
    # Properties #
    ##############

    XML_NAME = "pool"
    _XML_PROP_ORDER = [
        "name",
        "type",
        "uuid",
        "capacity",
        "allocation",
        "available",
        "format",
        "hosts",
        "_source_dir",
        "_source_adapter",
        "_source_device",
        "source_name",
        "target_path",
        "permissions",
        "auth_type",
        "auth_username",
        "auth_secret_uuid",
    ]

    _source_dir = XMLProperty("./source/dir/@path")
    _source_adapter = XMLProperty("./source/adapter/@name")
    _source_device = XMLProperty("./source/device/@path")

    type = XMLProperty("./@type")
    uuid = XMLProperty("./uuid")

    capacity = XMLProperty("./capacity", is_int=True)
    allocation = XMLProperty("./allocation", is_int=True)
    available = XMLProperty("./available", is_int=True)

    format = XMLProperty("./source/format/@type")
    iqn = XMLProperty("./source/initiator/iqn/@name")
    source_name = XMLProperty("./source/name")

    auth_type = XMLProperty("./source/auth/@type")
    auth_username = XMLProperty("./source/auth/@username")
    auth_secret_uuid = XMLProperty("./source/auth/secret/@uuid")

    target_path = XMLProperty("./target/path")

    hosts = XMLChildProperty(_Host, relative_xpath="./source")

    ######################
    # Public API helpers #
    ######################

    def supports_target_path(self):
        return self.type in [
            self.TYPE_DIR,
            self.TYPE_FS,
            self.TYPE_NETFS,
            self.TYPE_ISCSI,
            self.TYPE_SCSI,
            self.TYPE_MPATH,
        ]

    def supports_source_name(self):
        return self.type in [self.TYPE_LOGICAL, self.TYPE_GLUSTER, self.TYPE_RBD, self.TYPE_ZFS]

    def supports_source_path(self):
        return self.type in [
            self.TYPE_FS,
            self.TYPE_NETFS,
            self.TYPE_DISK,
            self.TYPE_ISCSI,
            self.TYPE_SCSI,
            self.TYPE_GLUSTER,
        ]

    def supports_hosts(self):
        return self.type in [self.TYPE_NETFS, self.TYPE_ISCSI, self.TYPE_GLUSTER, self.TYPE_RBD]

    def supports_format(self):
        return self.type in [self.TYPE_FS, self.TYPE_NETFS, self.TYPE_DISK]

    def supports_iqn(self):
        return self.type in [self.TYPE_ISCSI]

    def get_disk_type(self):
        if (
            self.type == StoragePool.TYPE_DISK
            or self.type == StoragePool.TYPE_LOGICAL
            or self.type == StoragePool.TYPE_SCSI
            or self.type == StoragePool.TYPE_MPATH
            or self.type == StoragePool.TYPE_ZFS
        ):
            return StorageVolume.TYPE_BLOCK
        if (
            self.type == StoragePool.TYPE_GLUSTER
            or self.type == StoragePool.TYPE_RBD
            or self.type == StoragePool.TYPE_ISCSI
        ):
            return StorageVolume.TYPE_NETWORK
        return StorageVolume.TYPE_FILE

    ##################
    # Build routines #
    ##################

    def validate(self):
        self.validate_name(self.conn, self.name)

        if not self.target_path:
            if self.type == self.TYPE_DISK:
                # disk is a bit special, in that it demands a target path,
                # but basically can't handle anything other than /dev
                self.target_path = _DEFAULT_DEV_TARGET
            else:
                self.target_path = self.default_target_path()
        if not self.source_name:
            self.source_name = self.default_source_name()
        if not self.format and self.supports_format():
            self.format = "auto"

        if self.type == self.TYPE_DISK and self.format == "auto":
            # There is no explicit "auto" type for disk pools, but leaving out
            # the format type seems to do the job for existing formatted disks
            self.format = None

    def install(self, meter=None, create=False, build=False, autostart=False):
        """
        Install storage pool xml.
        """
        xml = self.get_xml()
        log.debug("Creating storage pool '%s' with xml:\n%s", self.name, xml)

        if xmlutil.in_testsuite() and "virtinst-testsuite-fail-pool-install" in xml:
            raise RuntimeError("StoragePool.install testsuite mocked failure")

        meter = progress.ensure_meter(meter)

        try:
            pool = self.conn.storagePoolDefineXML(xml, 0)
        except Exception as e:  # pragma: no cover
            msg = _("Could not define storage pool: %s") % str(e)
            raise RuntimeError(msg) from None

        errmsg = None
        if build:
            try:
                pool.build(libvirt.VIR_STORAGE_POOL_BUILD_NEW)
            except Exception as e:  # pragma: no cover
                errmsg = _("Could not build storage pool: %s") % str(e)

        if create and not errmsg:
            try:
                pool.create(0)
            except Exception as e:  # pragma: no cover
                errmsg = _("Could not start storage pool: %s") % str(e)

        if autostart and not errmsg:
            try:
                pool.setAutostart(True)
            except Exception as e:  # pragma: no cover
                errmsg = _("Could not set pool autostart flag: %s") % str(e)

        if errmsg:  # pragma: no cover
            # Try and clean up the leftover pool
            try:
                pool.undefine()
            except Exception as e:
                log.debug("Error cleaning up pool after failure: %s", str(e))
            raise RuntimeError(errmsg)

        self.conn.cache_new_pool(pool)

        return pool


def _progress_thread(volname, pool, meter, event):
    vol = None
    if not meter:
        return  # pragma: no cover

    while True:
        try:
            if not vol:
                vol = pool.storageVolLookupByName(volname)
            vol.info()  # pragma: no cover
            break  # pragma: no cover
        except Exception:
            if event.wait(0.2):
                break

    if vol is None:
        log.debug("Couldn't lookup storage volume in prog thread.")
        return

    while True:  # pragma: no cover
        dummy1, dummy2, alloc = vol.info()
        meter.update(alloc)
        if event.wait(1):
            break


class StorageVolume(_StorageObject):
    """
    Base class for building and installing libvirt storage volume xml
    """

    @staticmethod
    def get_file_extension_for_format(fmt):
        if not fmt:
            return ""
        if fmt == "raw":
            return ".img"
        return "." + fmt

    @staticmethod
    def find_free_name(conn, pool_object, basename, collideguest=None, **kwargs):
        """
        Finds a name similar (or equal) to passed 'basename' that is not
        in use by another volume. Extra params are passed to generate_name

        :param collideguest: Guest object. If specified, also check to
        ensure we don't collide with any disk paths there
        """
        collidelist = []
        if collideguest:
            pooltarget = None
            poolname = pool_object.name()
            for poolxml in conn.fetch_all_pools():
                if poolxml.name == poolname:
                    pooltarget = poolxml.target_path
                    break

            for disk in collideguest.devices.disk:
                checkpath = disk.get_source_path()
                if pooltarget and checkpath and os.path.dirname(checkpath) == pooltarget:
                    collidelist.append(os.path.basename(checkpath))

        def cb(tryname):
            if tryname in collidelist:
                return True
            return generatename.check_libvirt_collision(pool_object.storageVolLookupByName, tryname)

        StoragePool.ensure_pool_is_running(pool_object, refresh=True)
        return generatename.generate_name(basename, cb, **kwargs)

    TYPE_FILE = getattr(libvirt, "VIR_STORAGE_VOL_FILE", 0)
    TYPE_BLOCK = getattr(libvirt, "VIR_STORAGE_VOL_BLOCK", 1)
    TYPE_DIR = getattr(libvirt, "VIR_STORAGE_VOL_DIR", 2)
    TYPE_NETWORK = getattr(libvirt, "VIR_STORAGE_VOL_NETWORK", 3)
    TYPE_NETDIR = getattr(libvirt, "VIR_STORAGE_VOL_NETDIR", 4)

    def __init__(self, *args, **kwargs):
        _StorageObject.__init__(self, *args, **kwargs)

        self._input_vol = None
        self._pool = None
        self._pool_xml = None
        self._reflink = False

    ######################
    # Non XML properties #
    ######################

    def _get_pool(self):
        return self._pool

    def _set_pool(self, newpool):
        StoragePool.ensure_pool_is_running(newpool)
        self._pool = newpool
        self._pool_xml = StoragePool(self.conn, parsexml=self._pool.XMLDesc(0))

    pool = property(_get_pool, _set_pool)

    @property
    def input_vol(self):
        return self._input_vol

    def set_input_vol(self, vol):
        self._input_vol = vol
        parsevol = StorageVolume(self.conn, parsexml=self._input_vol.XMLDesc(0))

        self.format = parsevol.format
        self.capacity = parsevol.capacity
        self.allocation = parsevol.allocation
        self.permissions.mode = parsevol.permissions.mode
        if not self._pool:
            self.pool = self._input_vol.storagePoolLookupByVolume()

    def _get_reflink(self):
        return self._reflink

    def _set_reflink(self, reflink):
        self._reflink = reflink

    reflink = property(_get_reflink, _set_reflink)

    ##########################
    # XML validation helpers #
    ##########################

    @staticmethod
    def validate_name(pool, name):
        XMLBuilder.validate_generic_name(_("Storage object"), name)

        try:
            pool.storageVolLookupByName(name)
        except libvirt.libvirtError:
            return
        raise ValueError(
            _("Name '%s' already in use by another volume.") % name
        )  # pragma: no cover

    def _get_vol_type(self):
        if self.type:  # pragma: no cover
            if self.type == "file":
                return self.TYPE_FILE
            elif self.type == "block":
                return self.TYPE_BLOCK
            elif self.type == "dir":
                return self.TYPE_DIR
            elif self.type == "network":
                return self.TYPE_NETWORK
        return self._pool_xml.get_disk_type()

    file_type = property(_get_vol_type)

    ##################
    # XML properties #
    ##################

    XML_NAME = "volume"
    _XML_PROP_ORDER = [
        "name",
        "key",
        "capacity",
        "allocation",
        "format",
        "target_path",
        "permissions",
    ]

    type = XMLProperty("./@type")
    key = XMLProperty("./key")
    capacity = XMLProperty("./capacity", is_int=True)
    allocation = XMLProperty("./allocation", is_int=True)
    format = XMLProperty("./target/format/@type")
    target_path = XMLProperty("./target/path")
    backing_store = XMLProperty("./backingStore/path")
    backing_format = XMLProperty("./backingStore/format/@type")
    lazy_refcounts = XMLProperty("./target/features/lazy_refcounts", is_bool=True)

    def _detect_backing_store_format(self):
        log.debug("Attempting to detect format for backing_store=%s", self.backing_store)
        from . import diskbackend

        path, vol, pool = diskbackend.manage_path(self.conn, self.backing_store)
        dummy = path

        if not vol:  # pragma: no cover
            log.debug("Didn't find any volume for backing_store")
            return None

        # Only set backing format for volumes that support
        # the 'format' parameter as we know it, like qcow2 etc.
        volxml = StorageVolume(self.conn, vol.XMLDesc(0))
        volxml.pool = pool
        log.debug("Found backing store volume XML:\n%s", volxml.get_xml())

        if not volxml.supports_format():  # pragma: no cover
            log.debug(
                "backing_store volume doesn't appear to have "
                "a file format we can specify, returning None"
            )
            return None

        log.debug("Returning format=%s", volxml.format)
        return volxml.format

    ######################
    # Public API helpers #
    ######################

    def supports_format(self):
        return self.file_type == self.TYPE_FILE

    ##################
    # Build routines #
    ##################

    def validate(self):
        self.validate_name(self.pool, self.name)

        if not self.format and self.file_type == self.TYPE_FILE:
            self.format = "raw"
        if self._prop_is_unset("lazy_refcounts") and self.format == "qcow2":
            self.lazy_refcounts = self.conn.support.conn_qcow2_lazy_refcounts()

        if self._pool_xml.type == StoragePool.TYPE_LOGICAL:
            if self.allocation != self.capacity:
                log.warning(
                    _(
                        "Sparse logical volumes are not supported, "
                        "setting allocation equal to capacity"
                    )
                )
                self.allocation = self.capacity

        isfatal, errmsg = self.is_size_conflict()
        if isfatal:
            raise ValueError(errmsg)
        if errmsg:
            log.warning(errmsg)

    def install(self, meter=None):
        """
        Build and install storage volume from xml
        """
        if self.backing_store and not self.backing_format:
            self.backing_format = self._detect_backing_store_format()

        xml = self.get_xml()
        log.debug("Creating storage volume '%s' with xml:\n%s", self.name, xml)

        cloneflags = 0
        createflags = 0
        if (
            self.format == "qcow2"
            and not self.backing_store
            and self.conn.support.pool_metadata_prealloc(self.pool)
        ):
            createflags |= libvirt.VIR_STORAGE_VOL_CREATE_PREALLOC_METADATA
            if self.capacity == self.allocation:
                # For cloning, this flag will make libvirt+qemu-img preallocate
                # the new disk image
                cloneflags |= libvirt.VIR_STORAGE_VOL_CREATE_PREALLOC_METADATA

        if self.reflink:
            if self.format != "raw":
                log.warning("skipping reflink for non-raw vol=%s", self.name)
            else:
                cloneflags |= getattr(libvirt, "VIR_STORAGE_VOL_CREATE_REFLINK", 1)

        event = threading.Event()
        meter = progress.ensure_meter(meter)
        t = threading.Thread(
            target=_progress_thread,
            name="Checking storage allocation",
            args=(self.name, self.pool, meter, event),
        )
        t.daemon = True

        try:
            t.start()
            msg = _("Allocating '%(filename)s'") % {"filename": self.name}
            meter.start(msg, self.capacity)

            if self.conn.is_really_test():
                # Test suite doesn't support any flags, so reset them
                createflags = 0
                cloneflags = 0

            if self.input_vol:
                vol = self.pool.createXMLFrom(xml, self.input_vol, cloneflags)
            else:
                log.debug("Using vol create flags=%s", createflags)
                vol = self.pool.createXML(xml, createflags)

            meter.update(self.capacity)
            meter.end()
            log.debug("Storage volume '%s' install complete.", self.name)
            return vol
        except Exception as e:
            log.debug("Error creating storage volume", exc_info=True)
            msg = "Couldn't create storage volume '%s': '%s'" % (self.name, str(e))
            raise RuntimeError(msg) from None
        finally:
            event.set()
            t.join()

    def is_size_conflict(self):
        """
        Report if requested size exceeds its pool's available amount

        :returns: 2 element tuple:
            1. True if collision is fatal, false otherwise
            2. String message if some collision was encountered.
        """
        if not self.pool:
            return (False, "")

        # pool info is [pool state, capacity, allocation, available]
        avail = self.pool.info()[3]
        if self.allocation > avail:
            msg = _(
                "There is not enough free space on the storage "
                "pool to create the volume. (%(mem1)s M requested "
                "allocation > %(mem2)s M available)"
            ) % {"mem1": (self.allocation // (1024 * 1024)), "mem2": (avail // (1024 * 1024))}
            return (True, msg)
        elif self.capacity > avail:
            msg = _(
                "The requested volume capacity will exceed the "
                "available pool space when the volume is fully "
                "allocated. (%(mem1)s M requested "
                "capacity > %(mem2)s M available)"
            ) % {"mem1": (self.capacity // (1024 * 1024)), "mem2": (avail // (1024 * 1024))}
            return (False, msg)
        return (False, "")



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .device import Device
from ..xmlbuilder import XMLProperty


class DevicePstore(Device):
    XML_NAME = "pstore"

    backend = XMLProperty("./@backend")
    path = XMLProperty("./path")
    size = XMLProperty("./size", is_int=True)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import tempfile

import libvirt
import pytest

import virtinst
from virtinst import log

import tests
from . import lib


def _vm_wrapper(vmname, uri="qemu:///system", opts=None):
    """
    Decorator to define+start a VM and clean it up on exit
    """

    def wrap1(fn):
        def wrapper(app, *args, **kwargs):
            app.error_if_already_running()
            xmlfile = "%s/live/%s.xml" % (tests.utils.UITESTDATADIR, vmname)
            conn = libvirt.open(uri)
            dom = conn.defineXML(open(xmlfile).read())
            try:
                dom.create()
                app.uri = uri
                app.conn = conn
                extra_opts = opts or []
                extra_opts += ["--show-domain-console", vmname]
                # Enable stats for more code coverage
                keyfile = "statsonly.ini"
                app.open(extra_opts=extra_opts, keyfile=keyfile)
                fn(app, dom, *args, **kwargs)
            finally:
                try:
                    app.stop()
                except Exception:
                    pass
                try:
                    flags = 0
                    if "qemu" in uri:
                        flags = (
                            libvirt.VIR_DOMAIN_UNDEFINE_NVRAM
                            | libvirt.VIR_DOMAIN_UNDEFINE_SNAPSHOTS_METADATA
                        )
                    dom.undefineFlags(flags)
                    dom.destroy()
                except Exception:
                    pass

        return wrapper

    return wrap1


def _create_qcow2_file(fn):
    def wrapper(app, *args, **kwargs):
        tmpdir = tempfile.TemporaryDirectory(prefix="uitests-tmp")
        dname = tmpdir.name
        try:
            fname = os.path.join(dname, "test.img")
            os.system("qemu-img create -f qcow2 %s 1M > /dev/null" % fname)
            os.system("chmod 700 %s" % dname)
            fn(fname, app, *args, **kwargs)
        finally:
            poolname = os.path.basename(dname)
            try:
                pool = app.conn.storagePoolLookupByName(poolname)
                pool.destroy()
                pool.undefine()
            except Exception:
                log.debug("Error cleaning up pool", exc_info=True)

    return wrapper


def _destroy(app, win):
    smenu = win.find("Menu", "toggle button")
    smenu.click()
    smenu.find("Force Off", "menu item").click()
    app.click_alert_button("you sure", "Yes")
    run = win.find("Run", "push button")
    lib.utils.check(lambda: run.sensitive)


###############################################
# Test live console connections with stub VMs #
###############################################


def _checkConsoleStandard(app, dom):
    """
    Shared logic for general console handling
    """
    ignore = dom
    win = app.topwin
    con = win.find("console-gfx-viewport")
    lib.utils.check(lambda: con.showing)

    win.find("Virtual Machine", "menu").click()
    win.find("Take Screenshot", "menu item").click()
    chooser = app.root.find(None, "file chooser")
    fname = chooser.find("Name", "text").text
    app.rawinput.pressKey("Enter")
    lib.utils.check(lambda: os.path.exists(fname))
    os.unlink(fname)
    lib.utils.check(lambda: win.active)

    win.find("Send Key", "menu").click()
    win.find(r"Ctrl\+Alt\+F1", "menu item").click()
    win.find("Send Key", "menu").click()
    win.find(r"Ctrl\+Alt\+F10", "menu item").click()
    win.find("Send Key", "menu").click()
    win.find(r"Ctrl\+Alt\+Delete", "menu item").click()

    # 'Resize to VM' testing
    oldsize = win.size
    win.find("^View$", "menu").click()
    scalemenu = win.find("Scale Display", "menu")
    scalemenu.point()
    scalemenu.find("Never", "radio menu item").click()
    win.find("^View$", "menu").click()
    win.find("Resize to VM", "menu item").click()
    newsize = win.size
    lib.utils.check(lambda: oldsize != newsize)

    # Fullscreen testing
    win.find("^View$", "menu").click()
    win.find("Fullscreen", "check menu item").click()
    fstb = win.find("Fullscreen Toolbar")
    lib.utils.check(lambda: fstb.showing)
    lib.utils.check(lambda: win.size != newsize)

    # Wait for toolbar to hide, then reveal it again
    lib.utils.check(lambda: not fstb.showing, timeout=5)
    app.rawinput.point(win.position[0] + win.size[0] / 2, 0)
    lib.utils.check(lambda: fstb.showing)
    # Move it off and have it hide again
    win.point()
    lib.utils.check(lambda: not fstb.showing, timeout=5)
    app.rawinput.point(win.position[0] + win.size[0] / 2, 0)
    lib.utils.check(lambda: fstb.showing)

    # Click stuff and exit fullscreen
    win.find("Fullscreen Send Key").click()
    app.rawinput.pressKey("Escape")
    win.find("Fullscreen Exit").click()
    lib.utils.check(lambda: win.size == newsize)

    # Trigger pointer grab, verify title was updated
    win.click()
    lib.utils.check(lambda: "Control_L" in win.name)
    # Ungrab
    win.keyCombo("<ctrl><alt>")
    lib.utils.check(lambda: "Control_L" not in win.name)

    # Tweak scaling
    win.window_maximize()
    win.find("^View$", "menu").click()
    scalemenu = win.find("Scale Display", "menu")
    scalemenu.point()
    scalemenu.find("Only", "radio menu item").click()
    win.find("^View$", "menu").click()
    scalemenu = win.find("Scale Display", "menu")
    scalemenu.point()
    scalemenu.find("Always", "radio menu item").click()

    # 'Resize to VM' again, to hit the scaling->always case
    oldsize = win.size
    win.find("^View$", "menu").click()
    win.find("Resize to VM", "menu item").click()
    newsize = win.size
    lib.utils.check(lambda: oldsize != newsize)

    win.window_close()


@_vm_wrapper("uitests-vnc-standard")
def testConsoleVNCStandard(app, dom):
    return _checkConsoleStandard(app, dom)


@_vm_wrapper("uitests-spice-standard")
def testConsoleSpiceStandard(app, dom):
    return _checkConsoleStandard(app, dom)


def _checkConsoleFocus(app, dom):
    """
    Shared logic for console keyboard grab handling
    """
    win = app.topwin
    con = win.find("console-gfx-viewport")
    lib.utils.check(lambda: con.showing)

    # Check that modifiers don't work when console grabs pointer
    win.click()
    app.sleep(0.5)  # make sure window code has time to adjust modifiers
    win.keyCombo("<ctrl><shift>w")
    lib.utils.check(lambda: win.showing)
    dom.destroy()
    win.find("Guest is not running.")
    win.grab_focus()
    app.sleep(0.5)  # make sure window code has time to adjust modifiers
    win.keyCombo("<ctrl><shift>w")
    lib.utils.check(lambda: not win.showing)


@_vm_wrapper("uitests-vnc-standard")
def testConsoleVNCFocus(app, dom):
    return _checkConsoleFocus(app, dom)


@_vm_wrapper("uitests-spice-standard")
def testConsoleSpiceFocus(app, dom):
    return _checkConsoleFocus(app, dom)


def _checkPassword(app):
    """
    Shared logic for password handling
    """
    win = app.topwin
    con = win.find("console-gfx-viewport")
    lib.utils.check(lambda: not con.showing)
    passwd = win.find("Password:", "password text")
    lib.utils.check(lambda: passwd.showing)

    # Check wrong password handling
    passwd.typeText("xx")
    win.find("Login", "push button").click()
    app.click_alert_button("Viewer authentication error", "OK")
    savecheck = win.find("Save this password", "check box")
    if not savecheck.checked:
        savecheck.click()
    passwd.typeText("yy")
    app.rawinput.pressKey("Enter")
    app.click_alert_button("Viewer authentication error", "OK")

    # Check proper password
    passwd.text = ""
    passwd.typeText("goodp")
    win.find("Login", "push button").click()
    lib.utils.check(lambda: con.showing)

    # Restart VM to retrigger console connect
    _destroy(app, win)
    win.find("Run", "push button").click()
    lib.utils.check(lambda: passwd.showing)
    # Password should be filled in
    lib.utils.check(lambda: bool(passwd.text))
    # Uncheck 'Save password' and login, which will delete it from keyring
    savecheck.click()
    win.find("Login", "push button").click()
    lib.utils.check(lambda: con.showing)

    # Restart VM to retrigger console connect
    _destroy(app, win)
    win.find("Run", "push button").click()
    lib.utils.check(lambda: passwd.showing)
    # Password should be empty now
    lib.utils.check(lambda: not bool(passwd.text))


@_vm_wrapper("uitests-vnc-password")
def testConsoleVNCPassword(app, dom):
    ignore = dom
    return _checkPassword(app)


@_vm_wrapper("uitests-spice-password")
def testConsoleSpicePassword(app, dom):
    ignore = dom
    return _checkPassword(app)


@_vm_wrapper("uitests-vnc-password", opts=["--test-options=fake-vnc-username"])
def testConsoleVNCPasswordUsername(app, dom):
    ignore = dom
    win = app.topwin
    con = win.find("console-gfx-viewport")
    lib.utils.check(lambda: not con.showing)
    passwd = win.find("Password:", "password text")
    lib.utils.check(lambda: passwd.showing)
    username = win.find("Username:", "text")
    lib.utils.check(lambda: username.showing)

    # Since we are mocking the username, sending the credentials
    # is ignored, so with the correct password this succeeds
    username.text = "fakeuser"
    passwd.typeText("goodp")
    win.find("Login", "push button").click()
    lib.utils.check(lambda: con.showing)


@_vm_wrapper("uitests-vnc-socket")
def testConsoleVNCSocket(app, dom):
    ignore = dom
    win = app.topwin
    con = win.find("console-gfx-viewport")
    lib.utils.check(lambda: con.showing)

    def _click_textconsole_menu(msg):
        vmenu = win.find("^View$", "menu")
        vmenu.click()
        tmenu = win.find("Consoles", "menu")
        tmenu.point()
        app.sleep(0.5)  # give console menu time to dynamically populate
        tmenu.find(msg, "radio menu item").click()

    # A bit of an extra test, make sure selecting Graphical Console works
    _click_textconsole_menu("Serial 1")
    lib.utils.check(lambda: not con.showing)
    _click_textconsole_menu("Graphical Console")
    lib.utils.check(lambda: con.showing)


def _testConsoleAutoconnect(app, dom, wname):
    ignore = dom
    win = app.topwin
    con = win.find(wname)
    lib.utils.check(lambda: con.showing)

    # Disable autoconnect
    vmenu = win.find("^View$", "menu")
    vmenu.click()
    vmenu.find("Autoconnect").click()
    dom.destroy()
    label = win.find("Guest is not running.")
    label.check_onscreen()
    dom.create()
    label.check_not_onscreen()
    button = win.find("Connect to console", "push button")
    button.check_onscreen()
    lib.utils.check(lambda: not con.showing)
    button.click()
    lib.utils.check(lambda: con.showing)


@_vm_wrapper("uitests-spice-standard")
def testConsoleAutoconnectGraphics(app, dom):
    _testConsoleAutoconnect(app, dom, "console-gfx-viewport")


@_vm_wrapper("uitests-lxc-serial", uri="lxc:///")
def testConsoleAutoconnectSerial(app, dom):
    _testConsoleAutoconnect(app, dom, "Serial Terminal")


@_vm_wrapper("uitests-lxc-serial", uri="lxc:///")
def testConsoleLXCSerial(app, dom):
    """
    Ensure LXC has serial open, and we can send some data
    """
    win = app.topwin
    term = win.find("Serial Terminal")
    lib.utils.check(lambda: term.showing)
    term.typeText("help\n")
    lib.utils.check(lambda: "COMMANDS" in term.text)

    term.doubleClick()
    term.click(button=3)
    menu = app.root.find("serial-popup-menu")
    menu.find("Copy", "menu item").click()

    term.click()
    term.click(button=3)
    menu = app.root.find("serial-popup-menu")
    menu.find("Paste", "menu item").click()

    win.find("Details", "radio button").click()
    win.find("Console", "radio button").click()
    _destroy(app, win)

    # Restart the guest to trigger reconnect code
    win.find("Run", "push button").click()
    term = win.find("Serial Terminal")
    lib.utils.check(lambda: term.showing)

    # Ensure ctrl+w doesn't close the window, modifiers are disabled
    term.click()
    win.keyCombo("<ctrl><shift>w")
    lib.utils.check(lambda: win.showing)
    # Shut it down, ensure accelerator works again
    _destroy(app, win)
    lib.utils.check(lambda: not dom.isActive())
    win.click_title()
    app.sleep(0.3)  # make sure window code has time to adjust modifiers
    win.keyCombo("<ctrl><shift>w")
    lib.utils.check(lambda: not win.showing)


@_vm_wrapper("uitests-spice-specific")
def testConsoleSpiceSpecific(app, dom):
    """
    Spice specific behavior. Has lots of devices that will open
    channels, spice GL + local config, and usbredir
    """
    ignore = dom
    win = app.topwin
    con = win.find("console-gfx-viewport")
    lib.utils.check(lambda: con.showing)

    # Just ensure the dialog pops up, can't really test much more
    # than that
    win.find("Virtual Machine", "menu").click()
    win.find("Redirect USB", "menu item").click()

    usbwin = app.root.find("vmm dialog", "alert")
    usbwin.find("Select USB devices for redirection", "label")
    usbwin.find("SPICE CD", "check box").click()
    chooser = app.root.find(None, "file chooser")
    # Find the cwd bookmark on the left
    chooser.find("virt-manager", "label").click()
    chooser.find("virt-manager", "label").click()
    chooser.find("COPYING").click()
    app.rawinput.pressKey("Enter")
    lib.utils.check(lambda: not chooser.showing)
    usbwin.find("Close", "push button").click()

    # Test fake guest resize behavior
    def _click_auto():
        vmenu = win.find("^View$", "menu")
        vmenu.click()
        smenu = vmenu.find("Scale Display", "menu")
        smenu.point()
        smenu.find("Auto resize VM", "check menu item").click()

    _click_auto()
    win.click_title()
    win.window_maximize()
    _click_auto()
    win.click_title()
    win.click_title()


@_vm_wrapper("uitests-vnc-standard")
def testVNCSpecific(app, dom):
    from gi.repository import GtkVnc

    if not hasattr(GtkVnc.Display, "set_allow_resize"):
        pytest.skip("GtkVnc is too old")

    ignore = dom
    win = app.topwin
    con = win.find("console-gfx-viewport")
    lib.utils.check(lambda: con.showing)

    # Test guest resize behavior
    def _click_auto():
        vmenu = win.find("^View$", "menu")
        vmenu.click()
        smenu = vmenu.find("Scale Display", "menu")
        smenu.point()
        smenu.find("Auto resize VM", "check menu item").click()

    _click_auto()
    win.click_title()
    win.window_maximize()
    _click_auto()
    win.click_title()
    win.click_title()


@_vm_wrapper("uitests-hotplug")
@_create_qcow2_file
def testLiveHotplug(fname, app, dom):
    """
    Live test for basic hotplugging and media change, as well as
    testing our auto-poolify magic
    """
    ignore = dom
    win = app.topwin
    win.find("Details", "radio button").click()

    # Add a scsi disk, importing the passed path
    win.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    addhw.find("Storage", "table cell").click()
    tab = addhw.find("storage-tab", None)
    lib.utils.check(lambda: tab.showing)
    tab.find("Select or create", "radio button").click()
    tab.find("storage-entry").set_text(fname)
    tab.combo_select("Bus type:", "SCSI")
    addhw.find("Finish", "push button").click()

    # Verify permission dialog pops up, ask to change
    app.click_alert_button("The emulator may not have search permissions", "Yes")

    # Verify no errors
    lib.utils.check(lambda: not addhw.showing)
    lib.utils.check(lambda: win.active)

    # Hot unplug the disk
    win.find("SCSI Disk 1", "table cell").click()
    tab = win.find("disk-tab", None)
    lib.utils.check(lambda: tab.showing)
    win.find("config-remove").click()
    delete = app.find_window("Remove Disk")
    delete.find_fuzzy("Delete", "button").click()
    lib.utils.check(lambda: not delete.active)
    lib.utils.check(lambda: os.path.exists(fname))

    # Change CDROM
    win.find("IDE CDROM 1", "table cell").click()
    tab = win.find("disk-tab", None)
    entry = win.find("media-entry")
    appl = win.find("config-apply")
    lib.utils.check(lambda: tab.showing)
    entry.set_text(fname)
    appl.click()

    lib.utils.check(lambda: not appl.sensitive)
    lib.utils.check(lambda: entry.text == fname)
    entry.click_secondary_icon()

    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    lib.utils.check(lambda: not entry.text)


@_vm_wrapper("uitests-hotplug")
@_create_qcow2_file
def testLiveExternalSnapshots(fname, app, dom):
    win = app.topwin
    win.find("Details", "radio button").click()

    # Add a scsi disk, importing the passed path
    win.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    addhw.find("Storage", "table cell").click()
    tab = addhw.find("storage-tab", None)
    lib.utils.check(lambda: tab.showing)
    tab.find("Select or create", "radio button").click()
    tab.find("storage-entry").set_text(fname)
    tab.combo_select("Bus type:", "SCSI")
    addhw.find("Finish", "push button").click()

    # Verify permission dialog pops up, ask to change
    app.click_alert_button("The emulator may not have search permissions", "Yes")

    # Verify no errors
    lib.utils.check(lambda: not addhw.showing)
    lib.utils.check(lambda: win.active)

    def _make_snapshot(name, auto=True, do_external=True):
        win.find("snapshot-add", "push button").click()
        newwin = app.find_window("Create snapshot")
        newwin.find("Name:", "text").set_text(name)
        external = newwin.find("external", "radio button")
        if not external.isChecked:
            pytest.skip("libvirt is too old for external snapshots")
        if not do_external:
            newwin.find("internal", "radio button").click()
        if not auto:
            newwin.find("auto", "check box").click()
        newwin.find("Finish", "push button").click()
        if not do_external:
            app.click_alert_button("Mixing external and internal snapshots", "No")
            newwin.find("Finish", "push button").click()
            app.click_alert_button("Mixing external and internal snapshots", "Yes")
        lib.utils.check(lambda: not newwin.showing)
        newc = win.find(name, "table cell")
        lib.utils.check(lambda: newc.state_selected)

    def _delete_snapshot(name):
        newc = win.find(name, "table cell")
        newc.click()
        lib.utils.check(lambda: newc.state_selected)
        win.find("snapshot-delete").click()
        app.click_alert_button("permanently delete", "Yes")
        lib.utils.check(lambda: newc.dead, timeout=10)
        lib.utils.check(lambda: win.active)

    win.find("Snapshots", "radio button").click()
    _make_snapshot("testnewsnap1")
    _make_snapshot("testnewsnap2", auto=False)

    # Poweroff VM and create an offline one
    run = win.find("Run", "push button")
    dom.destroy()
    lib.utils.check(lambda: run.sensitive)

    _make_snapshot("testnewsnap-offline")

    # Delete first snapshot
    _delete_snapshot("testnewsnap1")

    # Ensure VM is still offline
    lib.utils.check(lambda: run.sensitive)

    # Mix internal and external snapshots
    _make_snapshot("testnewsnap3", do_external=False)

    # Ensure newvm window defaults to internal now
    win.find("snapshot-add", "push button").click()
    newwin2 = app.find_window("Create snapshot")
    lib.utils.check(lambda: newwin2.find("internal", "radio button").isChecked)
    newwin2.find("Cancel", "push button").click()
    lib.utils.check(lambda: not newwin2.showing)

    # Delete snapshot and check the default reverts to external
    _delete_snapshot("testnewsnap3")
    win.find("snapshot-add", "push button").click()
    newwin2 = app.find_window("Create snapshot")
    lib.utils.check(lambda: newwin2.find("external", "radio button").isChecked)


@_vm_wrapper("uitests-firmware-efi")
def testFirmwareRename(app, dom):
    from virtinst import cli, DeviceDisk

    win = app.topwin
    dom.destroy()

    # First we refresh the 'nvram' pool, so we can reliably
    # check if nvram files are created/deleted as expected
    conn = cli.getConnection(app.conn.getURI())
    guest = virtinst.Guest(conn, dom.XMLDesc(0))
    origname = dom.name()
    origpath = guest.os.nvram
    if not origpath:
        pytest.skip("libvirt is too old to put nvram path in inactive XML")
    nvramdir = os.path.dirname(origpath)

    fakedisk = DeviceDisk(conn)
    fakedisk.set_source_path(nvramdir + "/FAKE-UITEST-FILE")
    nvram_pool = fakedisk.get_parent_pool()
    nvram_pool.refresh()

    newname = "uitests-firmware-efi-renamed"
    newpath = origpath.replace(origname + "_VARS", newname + "_VARS")
    assert DeviceDisk.path_definitely_exists(app.conn, origpath)
    assert not DeviceDisk.path_definitely_exists(app.conn, newpath)

    # Now do the actual UI clickage
    win.find("Details", "radio button").click()
    win.find("Hypervisor Details", "label")
    win.find("Overview", "table cell").click()

    newname = "uitests-firmware-efi-renamed"
    win.find("Name:", "text").set_text(newname)
    appl = win.find("config-apply")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Confirm window was updated
    app.find_window("%s on" % newname)

    # Confirm nvram paths were altered as expected
    assert not DeviceDisk.path_definitely_exists(app.conn, origpath)
    assert DeviceDisk.path_definitely_exists(app.conn, newpath)



# Copyright (C) 2006, 2013 Red Hat, Inc.
# Copyright (C) 2006 Daniel P. Berrange <berrange@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from gi.repository import Gio
from gi.repository import GLib

from virtinst import log

from ..baseclass import vmmGObject


class _vmmSecret:
    def __init__(self, name, secret=None, attributes=None):
        self.name = name
        self.secret = secret
        self.attributes = attributes

    def get_secret(self):
        return self.secret

    def get_name(self):
        return self.name


class vmmKeyring(vmmGObject):
    """
    freedesktop Secret API abstraction
    """

    @classmethod
    def get_instance(cls):
        if not cls._instance:
            cls._instance = vmmKeyring()
        return cls._instance

    def __init__(self):
        vmmGObject.__init__(self)

        self._collection = None

        try:
            self._dbus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
            self._service = Gio.DBusProxy.new_sync(
                self._dbus,
                0,
                None,
                "org.freedesktop.secrets",
                "/org/freedesktop/secrets",
                "org.freedesktop.Secret.Service",
                None,
            )

            self._session = self._service.OpenSession("(sv)", "plain", GLib.Variant("s", ""))[1]

            self._collection = Gio.DBusProxy.new_sync(
                self._dbus,
                0,
                None,
                "org.freedesktop.secrets",
                "/org/freedesktop/secrets/aliases/default",
                "org.freedesktop.Secret.Collection",
                None,
            )

            log.debug("Using keyring session %s", self._session)
        except Exception:  # pragma: no cover
            log.exception("Error determining keyring")

    def _cleanup(self):
        pass  # pragma: no cover

    def _find_secret_item_path(self, uuid, hvuri):
        attributes = {
            "uuid": uuid,
            "hvuri": hvuri,
        }
        unlocked, locked = self._service.SearchItems("(a{ss})", attributes)
        if not unlocked:
            if locked:
                log.warning("Item found, but it's locked")  # pragma: no cover
            return None
        return unlocked[0]

    def _do_prompt_if_needed(self, path):
        if path == "/":
            return
        iface = Gio.DBusProxy.new_sync(  # pragma: no cover
            self._dbus,
            0,
            None,
            "org.freedesktop.secrets",
            path,
            "org.freedesktop.Secret.Prompt",
            None,
        )
        iface.Prompt("(s)", "")  # pragma: no cover

    def _add_secret(self, secret):
        try:
            props = {
                "org.freedesktop.Secret.Item.Label": GLib.Variant("s", secret.get_name()),
                "org.freedesktop.Secret.Item.Attributes": GLib.Variant("a{ss}", secret.attributes),
            }
            params = (
                self._session,
                [],
                [ord(v) for v in secret.get_secret()],
                "text/plain; charset=utf8",
            )
            replace = True

            dummy, prompt = self._collection.CreateItem("(a{sv}(oayays)b)", props, params, replace)
            self._do_prompt_if_needed(prompt)
        except Exception:  # pragma: no cover
            log.exception("Failed to add keyring secret")

    def _del_secret(self, uuid, hvuri):
        try:
            path = self._find_secret_item_path(uuid, hvuri)
            if path is None:
                return None

            iface = Gio.DBusProxy.new_sync(
                self._dbus,
                0,
                None,
                "org.freedesktop.secrets",
                path,
                "org.freedesktop.Secret.Item",
                None,
            )
            prompt = iface.Delete()
            self._do_prompt_if_needed(prompt)
        except Exception:  # pragma: no cover
            log.exception("Failed to delete keyring secret")

    def _get_secret(self, uuid, hvuri):
        ret = None
        try:
            path = self._find_secret_item_path(uuid, hvuri)
            if path is None:
                return None

            iface = Gio.DBusProxy.new_sync(
                self._dbus,
                0,
                None,
                "org.freedesktop.secrets",
                path,
                "org.freedesktop.Secret.Item",
                None,
            )

            secretbytes = iface.GetSecret("(o)", self._session)[2]
            label = iface.get_cached_property("Label").unpack().strip("'")
            dbusattrs = iface.get_cached_property("Attributes").unpack()

            secret = "".join([chr(c) for c in secretbytes])

            attrs = {}
            for key, val in dbusattrs.items():
                if key not in ["hvuri", "uuid"]:
                    continue
                attrs["%s" % key] = "%s" % val

            ret = _vmmSecret(label, secret, attrs)
        except Exception:  # pragma: no cover
            log.exception("Failed to get keyring secret uuid=%r hvuri=%r", uuid, hvuri)

        return ret

    ##############
    # Public API #
    ##############

    def is_available(self):
        return self._collection is not None

    def _get_secret_name(self, vm):
        return "vm-console-" + vm.get_uuid()

    def get_console_password(self, vm):
        if not self.is_available():
            return ("", "")  # pragma: no cover

        secret = self._get_secret(vm.get_uuid(), vm.conn.get_uri())
        if secret is None:
            return ("", "")  # pragma: no cover

        return (secret.get_secret(), vm.get_console_username() or "")

    def set_console_password(self, vm, password, username=""):
        if not self.is_available():
            return  # pragma: no cover

        secret = _vmmSecret(
            self._get_secret_name(vm), password, {"uuid": vm.get_uuid(), "hvuri": vm.conn.get_uri()}
        )
        vm.set_console_username(username)
        self._add_secret(secret)

    def del_console_password(self, vm):
        if not self.is_available():
            return  # pragma: no cover

        self._del_secret(vm.get_uuid(), vm.conn.get_uri())
        vm.del_console_username()



# Copyright (C) 2013 Red Hat, Inc.
#
# Copyright 2012
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .device import Device
from ..xmlbuilder import XMLProperty


class DeviceMemballoon(Device):
    XML_NAME = "memballoon"

    model = XMLProperty("./@model")
    autodeflate = XMLProperty("./@autodeflate", is_onoff=True)
    stats_period = XMLProperty("./stats/@period", is_int=True)
    freePageReporting = XMLProperty("./@freePageReporting", is_onoff=True)

    ##################
    # Default config #
    ##################

    def set_defaults(self, guest):
        if not self.model:
            self.model = "virtio"



# Copyright 2005-2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import argparse
import atexit
import os
import sys
import time
import select

import libvirt

import virtinst
from . import cli
from .cli import fail, fail_conflicting, print_stdout, print_stderr
from . import Network
from .guest import Guest
from .logger import log


##############################
# Validation utility helpers #
##############################

INSTALL_METHODS = "--location URL, --cdrom CD/ISO, --pxe, --import, --boot hd|cdrom|..."


def supports_pxe(guest):
    """
    Return False if we are pretty sure the config doesn't support PXE
    """
    for nic in guest.devices.interface:
        if nic.type == nic.TYPE_USER:
            continue
        if nic.type != nic.TYPE_VIRTUAL:
            return True

        try:
            netobj = nic.conn.networkLookupByName(nic.source)
            xmlobj = Network(nic.conn, parsexml=netobj.XMLDesc(0))
            return xmlobj.can_pxe()
        except Exception:  # pragma: no cover
            log.debug("Error checking if PXE supported", exc_info=True)
            return True

    return False


def check_cdrom_option_error(options):
    if options.cdrom_short and options.cdrom:
        fail_conflicting("-c", "--cdrom")

    if options.cdrom_short:
        if "://" in options.cdrom_short:
            fail(
                _(
                    "-c specified with what looks like a libvirt URI. "
                    "Did you mean to use --connect? If not, use --cdrom "
                    "instead"
                )
            )
        options.cdrom = options.cdrom_short


#################################
# Back compat option conversion #
#################################


def convert_old_printxml(options):
    if options.xmlstep:
        options.xmlonly = options.xmlstep
        del options.xmlstep


def convert_old_sound(options):
    if not options.sound:
        return
    for idx, dummy in enumerate(options.sound):
        if options.sound[idx] is None:
            options.sound[idx] = "default"


def convert_old_init(options):
    if not options.init:
        return
    if not options.boot:
        options.boot = [""]
    options.boot[-1] += ",init=%s" % options.init
    log.debug("Converted old --init to --boot %s", options.boot[-1])


def _do_convert_old_disks(options):
    paths = virtinst.xmlutil.listify(options.file_paths)
    sizes = virtinst.xmlutil.listify(options.disksize)

    def padlist(lst, padsize):
        lst = virtinst.xmlutil.listify(lst)
        lst.extend((padsize - len(lst)) * [None])
        return lst

    disklist = padlist(paths, max(0, len(sizes)))
    sizelist = padlist(sizes, len(disklist))

    opts = []
    for idx, path in enumerate(disklist):
        optstr = ""
        if path:
            optstr += "path=%s" % path
        if sizelist[idx]:
            if optstr:
                optstr += ","
            optstr += "size=%s" % sizelist[idx]
        if options.sparse is False:
            if optstr:
                optstr += ","
            optstr += "sparse=no"
        log.debug("Converted to new style: --disk %s", optstr)
        opts.append(optstr)

    options.disk = opts


def convert_old_disks(options):
    if options.nodisks and (options.file_paths or options.disk or options.disksize):
        fail(_("Cannot specify storage and use --nodisks"))

    if (options.file_paths or options.disksize or not options.sparse) and options.disk:
        fail(
            _(
                "Cannot mix --file, --nonsparse, or --file-size with --disk "
                "options. Use --disk PATH[,size=SIZE][,sparse=yes|no]"
            )
        )

    if not options.disk:
        if options.nodisks:
            options.disk = ["none"]
        else:
            _do_convert_old_disks(options)

    del options.file_paths
    del options.disksize
    del options.sparse
    del options.nodisks
    log.debug("Distilled --disk options: %s", options.disk)


def convert_old_os_options(options):
    if not options.old_os_type:
        return
    log.warning(_("--os-type is deprecated and does nothing. Please stop using it."))
    del options.old_os_type


def convert_old_memory(options):
    if options.memory:
        return
    if not options.oldmemory:
        return
    options.memory = str(options.oldmemory)


def convert_old_cpuset(options):
    if not options.cpuset:
        return

    newvcpus = options.vcpus or []
    newvcpus.append(",cpuset=%s" % options.cpuset)
    options.vcpus = newvcpus
    log.debug("Generated compat cpuset: --vcpus %s", options.vcpus[-1])


def convert_old_networks(options):
    if options.nonetworks:
        options.network = ["none"]

    macs = virtinst.xmlutil.listify(options.mac)
    networks = virtinst.xmlutil.listify(options.network)
    bridges = virtinst.xmlutil.listify(options.bridge)

    if bridges and networks:
        fail_conflicting("--bridge", "--network")

    if bridges:
        # Convert old --bridges to --networks
        networks = ["bridge:" + b for b in bridges]

    def padlist(lst, padsize):
        lst = virtinst.xmlutil.listify(lst)
        lst.extend((padsize - len(lst)) * [None])
        return lst

    # If a plain mac is specified, have it imply a default network
    networks = padlist(networks, max(len(macs), 1))
    macs = padlist(macs, len(networks))

    for idx, ignore in enumerate(networks):
        if networks[idx] is None:
            networks[idx] = "default"
        if macs[idx]:
            networks[idx] += ",mac=%s" % macs[idx]

        # Handle old format of bridge:foo instead of bridge=foo
        for prefix in ["network", "bridge"]:
            if networks[idx].startswith(prefix + ":"):
                networks[idx] = networks[idx].replace(prefix + ":", prefix + "=")

    del options.mac
    del options.bridge
    del options.nonetworks

    options.network = networks
    log.debug("Distilled --network options: %s", options.network)


def convert_old_graphics(options):
    vnc = options.vnc
    vncport = options.vncport
    vnclisten = options.vnclisten
    nographics = options.nographics
    sdl = options.sdl
    keymap = options.keymap
    graphics = options.graphics

    if graphics and (vnc or sdl or keymap or vncport or vnclisten):
        fail(_("Cannot mix --graphics and old style graphical options"))

    optnum = sum(bool(g) for g in [vnc, nographics, sdl, graphics])
    if optnum > 1:
        raise ValueError(_("Can't specify more than one of VNC, SDL, --graphics or --nographics"))

    if options.graphics:
        return

    if optnum == 0:
        return

    # Build a --graphics command line from old style opts
    optstr = (vnc and "vnc") or (sdl and "sdl") or (nographics and ("none"))
    if vnclisten:
        optstr += ",listen=%s" % vnclisten
    if vncport:
        optstr += ",port=%s" % vncport
    if keymap:
        optstr += ",keymap=%s" % keymap

    log.debug("--graphics compat generated: %s", optstr)
    options.graphics = [optstr]


def convert_old_features(options):
    if options.features:
        return

    opts = []
    if options.noacpi:
        opts.append("acpi=off")
    if options.noapic:
        opts.append("apic=off")
    if opts:
        options.features = [",".join(opts)]


def convert_wait_zero(options):
    # Historical back compat, --wait 0 is identical to --noautoconsole
    if options.wait == 0:
        log.warning("Treating --wait 0 as --noautoconsole")
        options.autoconsole = "none"
        options.wait = None


##################################
# Install media setup/validation #
##################################


def do_test_media_detection(conn, options):
    url = options.test_media_detection
    guest = virtinst.Guest(conn)
    if options.arch:
        guest.os.arch = options.arch
    if options.os_type:
        guest.os.os_type = options.os_type
    guest.set_capabilities_defaults()

    installer = virtinst.Installer(conn, location=url)
    print_stdout(installer.detect_distro(guest), do_force=True)


#############################
# General option validation #
#############################


def storage_specified(options, guest):
    if guest.os.is_container():
        return True
    return options.disk or options.filesystem


def memory_specified(guest):
    return guest.memory or guest.currentMemory or guest.cpu.cells


def validate_required_options(options, guest, installer):
    # Required config. Don't error right away if nothing is specified,
    # aggregate the errors to help first time users get it right
    msg = ""

    if not options.reinstall:
        if not memory_specified(guest):
            msg += "\n" + _("--memory amount in MiB is required")

        if not storage_specified(options, guest):
            msg += "\n" + (_("--disk storage must be specified (override with --disk none)"))

    if not guest.os.is_container() and not installer.options_specified():
        msg += "\n" + (
            _("An install method must be specified\n(%(methods)s)") % {"methods": INSTALL_METHODS}
        )

    if msg:
        fail(msg)


def show_console_warnings(installer, autoconsole):
    if not installer.cdrom:
        return
    if not autoconsole.is_text():
        return
    log.warning(
        _(
            "CDROM media does not print to the text console "
            "by default, so you likely will not see text install output. "
            "You might want to use --location."
        )
        + " "
        + _("See the man page for examples of using --location with CDROM media")
    )


def _show_memory_warnings(guest):
    if not guest.currentMemory:
        return

    res = guest.osinfo.get_recommended_resources()
    rammb = guest.currentMemory // 1024
    minram = res.get_minimum_ram(guest.os.arch) or 0
    if minram:
        if (minram // 1024) > guest.currentMemory:
            log.warning(
                _(
                    "Requested memory %(mem1)s MiB is less than the "
                    "recommended %(mem2)s MiB for OS %(osname)s"
                ),
                {"mem1": rammb, "mem2": minram // (1024 * 1024), "osname": guest.osinfo.name},
            )
    elif rammb < 17:
        log.warning(
            _("Requested memory %s MiB is abnormally low. Were you trying to specify GiB?"),
            rammb,
        )


def _needs_accurate_osinfo(guest):
    # HVM is really the only case where OS impacts what we set for defaults,
    # so far.
    #
    # Historically we would only warn about missing osinfo on x86, but
    # with the change to make osinfo mandatory we relaxed the arch check,
    # so virt-install behavior is more consistent.
    return guest.os.is_hvm()


def show_guest_warnings(options, guest):
    if options.pxe and not supports_pxe(guest):
        log.warning(_("The guest's network configuration may not support PXE"))

    if guest.osinfo.is_generic() and _needs_accurate_osinfo(guest):
        log.warning(
            _(
                "Using --osinfo {osname}, VM performance may suffer. "
                "Specify an accurate OS for optimal results."
            ).format(osname=guest.osinfo.name)
        )

    _show_memory_warnings(guest)


##########################
# Guest building helpers #
##########################


def get_location_for_os(guest, osname, profile=None):
    osinfo = virtinst.OSDB.lookup_os(osname, raise_error=True)
    location = osinfo.get_location(guest.os.arch, profile)
    print_stdout(_("Using {osname} --location {url}").format(osname=osname, url=location))
    return location


def build_installer(options, guest, installdata):
    cdrom = None
    location = None
    location_kernel = None
    location_initrd = None
    is_reinstall = bool(options.reinstall)
    unattended_data = None
    extra_args = options.extra_args

    install_bootdev = installdata.bootdev
    install_kernel = installdata.kernel
    install_initrd = installdata.initrd
    install_kernel_args = installdata.kernel_args
    install_os = installdata.os
    no_install = installdata.no_install
    if installdata.kernel_args:
        if installdata.kernel_args_overwrite:
            install_kernel_args = installdata.kernel_args
        else:
            extra_args = [installdata.kernel_args]

    if options.unattended and options.cloud_init:
        cli.fail_conflicting("--unattended", "--cloud-init")

    if options.unattended:
        unattended_data = cli.parse_unattended(options.unattended)

    if install_os:
        profile = unattended_data.profile if unattended_data else None
        location = get_location_for_os(guest, install_os, profile)
    elif options.location:
        (location, location_kernel, location_initrd) = cli.parse_location(options.location)
    elif options.cdrom:
        cdrom = options.cdrom
        if options.livecd:
            no_install = True
    elif options.pxe:
        install_bootdev = "network"
    elif installdata.is_set:
        pass
    elif options.xmlonly:
        no_install = True
    elif options.import_install:
        no_install = True
    elif options.boot_was_set:
        no_install = True

    installer = virtinst.Installer(
        guest.conn,
        cdrom=cdrom,
        location=location,
        location_kernel=location_kernel,
        location_initrd=location_initrd,
        install_bootdev=install_bootdev,
        install_kernel=install_kernel,
        install_initrd=install_initrd,
        install_kernel_args=install_kernel_args,
        no_install=no_install,
        is_reinstall=is_reinstall,
    )

    if unattended_data:
        installer.set_unattended_data(unattended_data)
    if extra_args:
        installer.set_extra_args(extra_args)
    if options.initrd_inject:
        installer.set_initrd_injections(options.initrd_inject)
    if options.autostart:
        installer.autostart = True
    if options.cloud_init:
        cloudinit_data = cli.parse_cloud_init(options.cloud_init)
        installer.set_cloudinit_data(cloudinit_data)

    return installer


def set_cli_default_name(guest):
    if not guest.name:
        default_name = virtinst.Guest.generate_name(guest)
        cli.print_stdout(_("Using default --name {vm_name}").format(vm_name=default_name))
        guest.name = default_name


def set_cli_defaults(options, guest):
    if guest.os.is_container():
        if not memory_specified(guest):
            mbram = 1024
            # LXC doesn't even do anything with memory settings, but libvirt
            # XML requires it anyways. Fill in 64 MiB
            cli.print_stdout(
                _("Using container default --memory {megabytes}").format(megabytes=mbram)
            )
            guest.currentMemory = mbram * 1024
        return

    if (
        options.unattended
        and guest.osinfo.is_windows()
        and guest.osinfo.supports_unattended_drivers(guest.os.arch)
    ):
        guest.add_extra_drivers(guest.osinfo.get_pre_installable_devices(guest.os.arch))

    res = guest.osinfo.get_recommended_resources()
    storage = res.get_recommended_storage(guest.os.arch)
    ram = res.get_recommended_ram(guest.os.arch)
    ncpus = res.get_recommended_ncpus(guest.os.arch)

    if ram and not memory_specified(guest):
        mbram = str(ram / (1024 * 1024)).rstrip("0").rstrip(".")
        cli.print_stdout(
            _("Using {os_name} default --memory {megabytes}").format(
                os_name=guest.osinfo.name, megabytes=mbram
            )
        )
        guest.currentMemory = ram // 1024

    if ncpus:
        # We need to do this upfront, so we don't incorrectly set guest.vcpus
        guest.sync_vcpus_topology(ncpus)

    if storage and not storage_specified(options, guest):
        diskstr = "size=%d" % (storage // (1024**3))
        cli.print_stdout(
            _("Using {os_name} default --disk {disk_options}").format(
                os_name=guest.osinfo.name, disk_options=diskstr
            )
        )
        options.disk = [diskstr]
        cli.ParserDisk(diskstr, guest=guest).parse(None)


def set_explicit_guest_options(options, guest):
    if options.name:
        guest.name = options.name
        options.name = None
    if options.uuid:
        guest.uuid = options.uuid
        options.uuid = None
    if options.description:
        guest.description = options.description
        options.description = None
    if options.os_type:
        guest.os.os_type = options.os_type
        options.os_type = None
    if options.virt_type:
        guest.type = options.virt_type
        options.virt_type = None
    if options.arch:
        guest.os.arch = options.arch
        options.arch = None
    if options.machine:
        guest.os.machine = options.machine
        options.machine = None


def installer_detect_distro(guest, installer, osdata):
    os_set = False
    try:
        # OS name has to be set firstly whenever --osinfo is passed,
        # otherwise it won't be respected when the installer creates the
        # Distro Store.
        fallback_name = None
        if osdata.get_name():
            fallback_name = osdata.get_name()
            os_set = True
            guest.set_os_name(fallback_name)

        # This also validates the install location
        autodistro = installer.detect_distro(guest)
        if osdata.is_detect():
            if autodistro:
                os_set = True
                guest.set_os_name(autodistro)
            elif fallback_name:
                msg = _(
                    "Failed to detect osinfo OS name from install media, "
                    "using fallback name '{name}'.\n"
                    "Please file a bug against virt-install if "
                    "you expected the detection to succeed."
                ).format(name=fallback_name)
                log.warning(msg)
    except ValueError as e:
        fail(_("Error validating install location: %s") % str(e))

    msg = _("--osinfo/--os-variant OS name is required, but no value was\nset or detected.")
    if os_set:
        return
    if osdata.is_require_on():
        fail(msg)

    if not _needs_accurate_osinfo(guest):
        return

    fail_msg = msg + "\n\n"
    fail_msg += _(
        "This is now a fatal error. Specifying an OS name is required\n"
        "for modern, performant, and secure virtual machine defaults.\n"
    )

    detect_msg = _(
        "If you expected virt-install to detect an OS name from the\n"
        "install media, you can set a fallback OS name with:\n"
        "\n"
        "  --osinfo detect=on,name=OSNAME\n"
    )
    possibly_detectable = bool(installer.location or installer.cdrom)
    if possibly_detectable:
        fail_msg += "\n" + detect_msg

    fail_msg += "\n" + _(
        "You can see a full list of possible OS name values with:\n"
        "\n"
        "   virt-install --osinfo list\n"
    )

    generic_linux_names = [o.name for o in virtinst.OSDB.list_os() if o.is_linux_generic()]
    generic_linux_msg = _(
        "If your Linux distro is not listed, try one of generic values\nsuch as: {oslist}\n"
    ).format(oslist=", ".join(generic_linux_names))
    if generic_linux_names:
        fail_msg += "\n" + generic_linux_msg

    envkey = "VIRTINSTALL_OSINFO_DISABLE_REQUIRE"
    fail_msg += "\n" + _(
        "If you just need to get the old behavior back, you can use:\n"
        "\n"
        "  --osinfo detect=on,require=off\n"
        "\n"
        "Or export {env_var}=1\n"
    ).format(env_var=envkey)

    fail_msg = "\n" + fail_msg
    if envkey in os.environ:
        log.warning(fail_msg)
        m = _("{env_var} set. Skipping fatal error.").format(env_var=envkey)
        log.warning(m)
    else:
        fail(fail_msg)


def _build_options_guest(conn, options):
    guest = Guest(conn)
    guest.skip_default_osinfo = True

    # Fill in guest from the command line content
    set_explicit_guest_options(options, guest)

    # We do these two parser bit early, since Installer setup will
    # depend on them, but delay the rest to later, since things like
    # disk naming can depend on Installer operations
    cli.run_parser(guest, cli.ParserBoot, options.boot)
    cli.run_parser(guest, cli.ParserMetadata, options.metadata)

    # Call set_capabilities_defaults explicitly here rather than depend
    # on set_defaults calling it. Installer setup needs filled in values.
    # However we want to do it after run_all_parsers to ensure
    # we are operating on any arch/os/type values passed in with --boot
    guest.set_capabilities_defaults()

    return guest


def build_guest_instance(conn, options):
    installdata = cli.parse_install(options.install)
    osdata = cli.parse_osinfo(options.osinfo or installdata.os)
    options.boot_was_set = bool(options.boot)

    if options.reinstall:
        dummy1, guest, dummy2 = cli.get_domain_and_guest(conn, options.reinstall)
    else:
        guest = _build_options_guest(conn, options)

    installer = build_installer(options, guest, installdata)

    # Set guest osname, from commandline or detected from media
    guest.set_default_os_name()
    installer_detect_distro(guest, installer, osdata)

    if not options.reinstall:
        # We want to fill in --name before we do disk parsing, since
        # default disk paths are generated based on VM name
        set_cli_default_name(guest)
        cli.run_all_parsers(options, guest)
        set_cli_defaults(options, guest)

    installer.set_install_defaults(guest)
    for path in installer.get_search_paths(guest):
        cli.check_path_search(guest.conn, path)

    if not options.reinstall:
        # cli specific disk validation
        for disk in guest.devices.disk:
            cli.validate_disk(disk)
        for net in guest.devices.interface:
            cli.validate_mac(net.conn, net.macaddr)

    validate_required_options(options, guest, installer)
    show_guest_warnings(options, guest)

    return guest, installer


###########################
# Install process helpers #
###########################


def _sleep(secs):
    if not virtinst.xmlutil.in_testsuite():
        time.sleep(secs)  # pragma: no cover


def _set_default_wait(autoconsole, options):
    if options.wait is not None or autoconsole.has_console_cb() or not autoconsole.is_default():
        return

    # If there isn't any console to actually connect up,
    # default to --wait -1 to get similarish behavior
    log.warning(_("No console to launch for the guest, defaulting to --wait -1"))
    options.wait = -1


class WaitHandler:
    """
    Helper class for handling the --wait option sleeping and time tracking
    """

    def __init__(self, wait):
        self.wait_is_requested = False
        self._wait_mins = 0
        self._start_time = 0

        if wait is not None:
            self.wait_is_requested = True
            self._wait_mins = wait

    @property
    def wait_for_console_to_exit(self):
        # If --wait specified, we don't want the default behavior of waiting
        # for virt-viewer to exit, we want to launch it, then manually count
        # down time for ourselves
        return not self.wait_is_requested

    @property
    def _wait_forever(self):
        return self._wait_mins < 0

    @property
    def _wait_secs(self):
        return self._wait_mins * 60

    def start(self):
        self._start_time = time.time()

    def get_time_string(self):
        if self._wait_forever:
            return _("Waiting for the installation to complete.")
        return ngettext(
            "Waiting %(minutes)d minute for the installation to complete.",
            "Waiting %(minutes)d minutes for the installation to complete.",
            self._wait_mins,
        ) % {"minutes": self._wait_mins}

    def wait(self):
        """
        sleep 1 second, then return True if wait time has expired
        """
        _sleep(1)
        if self._wait_forever:
            if virtinst.xmlutil.in_testsuite():
                return True
            return False  # pragma: no cover

        time_elapsed = time.time() - self._start_time
        return (time_elapsed >= self._wait_secs) or virtinst.xmlutil.in_testsuite()


def _print_cloudinit_passwd(installer):
    passwd = installer.get_generated_password()
    if not passwd:
        return

    print_stdout(_("Password for first root login is: %s") % passwd, do_force=True, do_log=False)

    stdins = [sys.stdin]
    timeout = 10
    if sys.stdin.closed or not sys.stdin.isatty():
        if not virtinst.xmlutil.in_testsuite():  # pragma: no cover
            return
        stdins = []
        timeout = 0.0001

    sys.stdout.write(_("Installation will continue in 10 seconds (press Enter to skip)..."))
    sys.stdout.flush()

    select.select(stdins, [], [], timeout)


def _connect_console(guest, instdomain, autoconsole, wait):
    """
    Launched the passed console callback for the already defined
    domain. If domain isn't running, return an error.
    """
    console_cb = autoconsole.get_console_cb()
    if not console_cb:
        return

    child = console_cb(guest)
    if not wait:
        return

    # If we connected the console, wait for it to finish
    try:
        errcode = os.waitpid(child, 0)[1]
    except OSError as e:  # pragma: no cover
        log.debug("waitpid error: %s", e)

    if errcode:
        log.warning(_("Console command returned failure."))

    instdomain.handle_destroy_on_exit()


class _InstalledDomain:
    """
    Wrapper for the domain object after the initial install creation
    """

    def __init__(self, domain, transient, destroy_on_exit):
        self._domain = domain
        self._transient = transient
        self._destroy_on_exit = destroy_on_exit

        if destroy_on_exit:
            atexit.register(_destroy_on_exit, domain)

    def handle_destroy_on_exit(self):
        if self._destroy_on_exit and self._domain.isActive():
            log.debug("console exited and destroy_on_exit passed, destroying")
            self._domain.destroy()

    def domain_was_destroyed(self):
        try:
            state, reason = self._domain.state()
            return state == libvirt.VIR_DOMAIN_SHUTOFF and reason in [
                libvirt.VIR_DOMAIN_SHUTOFF_DESTROYED,
                libvirt.VIR_DOMAIN_SHUTOFF_SAVED,
            ]
        except Exception:  # pragma: no cover
            log.debug("Error checking VM shutdown reason", exc_info=True)

    def check_inactive(self):
        try:
            dominfo = self._domain.info()
            state = dominfo[0]

            if state == libvirt.VIR_DOMAIN_CRASHED:
                fail(_("Domain has crashed."))  # pragma: no cover

            return not self._domain.isActive()
        except libvirt.libvirtError as e:
            if e.get_error_code() != libvirt.VIR_ERR_NO_DOMAIN:
                raise  # pragma: no cover

            if self._transient:
                log.debug("transient VM shutdown and disappeared.")
                return True
            fail(_("VM disappeared unexpectedly: %s") % e)  # pragma: no cover


def _wait_for_domain(installer, instdomain, autoconsole, waithandler):
    """
    Make sure domain ends up in expected state, and wait if for install
    to complete if requested
    """
    if instdomain.check_inactive():
        return

    if bool(autoconsole.get_console_cb()):
        # We are trying to detect if the VM shutdown, or the user
        # just closed the console and the VM is still running. In the
        # the former case, libvirt may not have caught up yet with the
        # VM having exited, so wait a bit and check again
        _sleep(2)
        if instdomain.check_inactive():
            return  # pragma: no cover

    # If we reach here, the VM still appears to be running.
    msg = "\n"
    msg += _("Domain is still running. Installation may be in progress.")

    if not waithandler.wait_is_requested:
        # User either:
        #   used --noautoconsole
        #   killed console and guest is still running
        if not installer.requires_postboot_xml_changes():
            return

        msg += "\n"
        msg += _("You can reconnect to the console to complete the installation process.")
        print_stdout(msg)
        sys.exit(0)

    print_stdout(msg)
    print_stdout(waithandler.get_time_string())

    # Wait loop
    while True:
        if instdomain.check_inactive():  # pragma: no cover
            print_stdout(_("Domain has shutdown. Continuing."))
            break

        done = waithandler.wait()
        if done:
            print_stdout(_("Installation has exceeded specified time limit. Exiting application."))
            sys.exit(1)


def _testsuite_hack_destroy(domain):
    # Trigger specific behavior checking if user destroyed the domain
    if os.environ.get("VIRTINST_TESTSUITE_HACK_DESTROY"):
        domain.destroy()


def _process_domain(
    domain, guest, installer, waithandler, autoconsole, transient, destroy_on_exit, noreboot
):
    """
    Handle the pieces of the install process after the initial VM startup
    """
    instdomain = _InstalledDomain(domain, transient, destroy_on_exit)

    _connect_console(guest, instdomain, autoconsole, waithandler.wait_for_console_to_exit)

    _testsuite_hack_destroy(domain)
    _wait_for_domain(installer, instdomain, autoconsole, waithandler)
    print_stdout(_("Domain creation completed."))

    if transient:
        return
    if domain.isActive():
        return

    if noreboot or not installer.requires_postboot_xml_changes():
        print_stdout(  # pragma: no cover
            _("You can restart your domain by running:\n  %s") % cli.virsh_start_cmd(guest)
        )
        return

    if instdomain.domain_was_destroyed() and not destroy_on_exit:
        print_stdout(_("User stopped the VM. Not rebooting."))
        return

    print_stdout(_("Restarting guest."))
    domain.create()
    _connect_console(guest, instdomain, autoconsole, True)


def start_install(guest, installer, options):
    """
    Process all the install workflow specific options, and kick off
    the Installer process
    """
    autoconsole = cli.parse_autoconsole(options, guest, installer)
    show_console_warnings(installer, autoconsole)
    _set_default_wait(autoconsole, options)
    waithandler = WaitHandler(options.wait)
    meter = cli.get_meter()

    # we've got everything -- try to start the install
    print_stdout(_("\nStarting install..."))
    _print_cloudinit_passwd(installer)
    waithandler.start()

    try:
        try:
            domain = installer.start_install(
                guest, meter=meter, doboot=not options.noreboot, transient=options.transient
            )
        except:  # noqa: E722
            virtinst.Installer.cleanup_created_disks(guest, meter)
            raise

        _process_domain(
            domain,
            guest,
            installer,
            waithandler,
            autoconsole,
            options.transient,
            options.destroy_on_exit,
            options.noreboot,
        )

        if virtinst.xmlutil.in_testsuite() and options.destroy_on_exit:
            # Helps with unit testing
            _destroy_on_exit(domain)
    except KeyboardInterrupt:  # pragma: no cover
        log.debug("", exc_info=True)
        print_stderr(_("Domain install interrupted."))
        raise
    except Exception as e:
        fail(e, do_exit=False)
        cli.install_fail(guest)


########################
# XML printing helpers #
########################


def xml_to_print(guest, installer, xmlonly, dry):
    start_xml, final_xml = installer.start_install(guest, dry=dry, return_xml=True)
    if not start_xml:
        start_xml = final_xml
        final_xml = None

    if dry and not xmlonly:
        print_stdout(_("Dry run completed successfully"))
        return

    if xmlonly not in [False, "1", "2", "all"]:
        fail(_("Unknown XML step request '%s', must be 1, 2, or all") % xmlonly)

    if xmlonly == "1":
        return start_xml
    if xmlonly == "2":
        if not final_xml:
            fail(_("Requested installation does not have XML step 2"))
        return final_xml

    # "all" case
    xml = start_xml
    if final_xml:
        xml += final_xml
    return xml


#######################
# CLI option handling #
#######################


def parse_args():
    parser = cli.setupParser(
        "%(prog)s OPTIONS",
        _("Create a new virtual machine from specified install media."),
        introspection_epilog=True,
    )
    cli.add_connect_option(parser)

    geng = parser.add_argument_group(_("General Options"))
    geng.add_argument("-n", "--name", help=_("Name of the guest instance"))
    cli.add_memory_option(geng, backcompat=True)
    cli.vcpu_cli_options(geng)
    cli.add_metadata_option(geng)
    cli.add_xml_option(geng)
    geng.add_argument("-u", "--uuid", help=argparse.SUPPRESS)
    geng.add_argument("--description", help=argparse.SUPPRESS)

    insg = parser.add_argument_group(_("Installation Method Options"))
    insg.add_argument("-c", dest="cdrom_short", help=argparse.SUPPRESS)
    insg.add_argument("--cdrom", help=_("CD-ROM installation media"))
    insg.add_argument(
        "-l",
        "--location",
        help=_(
            "Distro install URL, eg. https://host/path. See man "
            "page for specific distro examples."
        ),
    )
    insg.add_argument(
        "--pxe", action="store_true", help=_("Boot from the network using the PXE protocol")
    )
    insg.add_argument(
        "--import",
        action="store_true",
        dest="import_install",
        help=_("Build guest around an existing disk image"),
    )
    insg.add_argument("--livecd", action="store_true", help=argparse.SUPPRESS)
    insg.add_argument(
        "-x",
        "--extra-args",
        action="append",
        help=_("Additional arguments to pass to the install kernel booted from --location"),
    )
    insg.add_argument(
        "--initrd-inject",
        action="append",
        help=_("Add given file to root of initrd from --location"),
    )
    insg.add_argument(
        "--unattended", nargs="?", const=1, help=_("Perform an unattended installation")
    )
    insg.add_argument("--install", help=_("Specify fine grained install options"))
    insg.add_argument(
        "--reinstall",
        metavar="DOMAIN",
        help=_(
            "Reinstall existing VM. Only install options are applied, "
            "all other VM configuration options are ignored."
        ),
    )
    insg.add_argument(
        "--cloud-init",
        nargs="?",
        const=1,
        help=_("Perform a cloud image installation, configuring cloud-init"),
    )

    # Takes a URL and just prints to stdout the detected distro name
    insg.add_argument("--test-media-detection", help=argparse.SUPPRESS)
    # Helper for cli testing, fills in standard stub options
    insg.add_argument("--test-stub-command", action="store_true", help=argparse.SUPPRESS)

    cli.add_boot_options(insg)
    insg.add_argument("--init", help=argparse.SUPPRESS)

    osg = cli.add_osinfo_option(parser, virtinstall=True)
    osg.add_argument("--os-type", dest="old_os_type", help=argparse.SUPPRESS)

    devg = parser.add_argument_group(_("Device Options"))
    cli.add_disk_option(devg)
    cli.add_net_option(devg)
    cli.add_gfx_option(devg)
    cli.add_device_options(devg, sound_back_compat=True)

    # Deprecated device options
    devg.add_argument("-f", "--file", dest="file_paths", action="append", help=argparse.SUPPRESS)
    devg.add_argument(
        "-s", "--file-size", type=float, action="append", dest="disksize", help=argparse.SUPPRESS
    )
    devg.add_argument(
        "--nonsparse", action="store_false", default=True, dest="sparse", help=argparse.SUPPRESS
    )
    devg.add_argument("--nodisks", action="store_true", help=argparse.SUPPRESS)
    devg.add_argument("--nonetworks", action="store_true", help=argparse.SUPPRESS)
    devg.add_argument("-b", "--bridge", action="append", help=argparse.SUPPRESS)
    devg.add_argument("-m", "--mac", action="append", help=argparse.SUPPRESS)
    devg.add_argument("--vnc", action="store_true", help=argparse.SUPPRESS)
    devg.add_argument("--vncport", type=int, help=argparse.SUPPRESS)
    devg.add_argument("--vnclisten", help=argparse.SUPPRESS)
    devg.add_argument("-k", "--keymap", help=argparse.SUPPRESS)
    devg.add_argument("--sdl", action="store_true", help=argparse.SUPPRESS)
    devg.add_argument("--nographics", action="store_true", help=argparse.SUPPRESS)

    gxmlg = parser.add_argument_group(_("Guest Configuration Options"))
    cli.add_guest_xml_options(gxmlg)

    virg = parser.add_argument_group(_("Virtualization Platform Options"))
    ostypeg = virg.add_mutually_exclusive_group()
    ostypeg.add_argument(
        "-v",
        "--hvm",
        action="store_const",
        const="hvm",
        dest="os_type",
        help=_("This guest should be a fully virtualized guest"),
    )
    ostypeg.add_argument(
        "-p",
        "--paravirt",
        action="store_const",
        const="xen",
        dest="os_type",
        help=_("This guest should be a paravirtualized guest"),
    )
    ostypeg.add_argument(
        "--container",
        action="store_const",
        const="exe",
        dest="os_type",
        help=_("This guest should be a container guest"),
    )
    virg.add_argument("--virt-type", help=_("Hypervisor name to use (kvm, qemu, xen, ...)"))
    virg.add_argument("--arch", help=_("The CPU architecture to simulate"))
    virg.add_argument("--machine", help=_("The machine type to emulate"))
    virg.add_argument("--accelerate", action="store_true", help=argparse.SUPPRESS)
    virg.add_argument("--noapic", action="store_true", default=False, help=argparse.SUPPRESS)
    virg.add_argument("--noacpi", action="store_true", default=False, help=argparse.SUPPRESS)

    misc = parser.add_argument_group(_("Miscellaneous Options"))
    misc.add_argument(
        "--autostart",
        action="store_true",
        default=False,
        help=_("Have domain autostart on host boot up."),
    )
    misc.add_argument(
        "--transient", action="store_true", default=False, help=_("Create a transient domain.")
    )
    misc.add_argument(
        "--destroy-on-exit",
        action="store_true",
        default=False,
        help=_("Force power off the domain when the console viewer is closed."),
    )
    misc.add_argument(
        "--wait", type=int, const=-1, nargs="?", help=_("Minutes to wait for install to complete.")
    )

    cli.add_misc_options(
        misc,
        prompt=True,
        printxml=True,
        printstep=True,
        noreboot=True,
        dryrun=True,
        noautoconsole=True,
    )

    cli.autocomplete(parser)

    return parser.parse_args()


###################
# main() handling #
###################


# Catchall for destroying the VM on ex. ctrl-c
def _destroy_on_exit(domain):
    try:
        isactive = bool(domain and domain.isActive())
        if isactive:
            domain.destroy()  # pragma: no cover
    except libvirt.libvirtError as e:  # pragma: no cover
        if e.get_error_code() != libvirt.VIR_ERR_NO_DOMAIN:
            log.debug("Error invoking atexit destroy_on_exit", exc_info=True)


def set_test_stub_options(options):  # pragma: no cover
    # Set some basic options that will let virt-install succeed. Helps
    # save boiler plate typing when testing new command line additions
    if not options.test_stub_command:
        return

    options.import_install = True
    if not options.connect:
        options.connect = "test:///default"
    if not options.name:
        options.name = "test-stub-command"
    if not options.memory:
        options.memory = "256"
    if not options.disk:
        options.disk = "none"
    if not options.graphics:
        options.graphics = "none"
    if not options.osinfo:
        options.osinfo = "fedora27"


def main(conn=None):
    cli.earlyLogging()
    options = parse_args()

    # Default setup options
    convert_old_printxml(options)
    options.quiet = options.xmlonly or options.test_media_detection or options.quiet
    cli.setupLogging("virt-install", options.debug, options.quiet)

    if cli.check_option_introspection(options):
        return 0
    if cli.check_osinfo_list(options):
        return 0

    check_cdrom_option_error(options)
    cli.convert_old_force(options)
    cli.parse_check(options.check)
    cli.set_prompt(options.prompt)
    convert_old_memory(options)
    convert_old_sound(options)
    convert_old_networks(options)
    convert_old_graphics(options)
    convert_old_disks(options)
    convert_old_features(options)
    convert_old_cpuset(options)
    convert_old_init(options)
    convert_wait_zero(options)
    set_test_stub_options(options)
    convert_old_os_options(options)

    conn = cli.getConnection(options.connect, conn=conn)

    if options.test_media_detection:
        do_test_media_detection(conn, options)
        return 0

    guest, installer = build_guest_instance(conn, options)
    if options.xmlonly or options.dry:
        xml = xml_to_print(guest, installer, options.xmlonly, options.dry)
        if xml:
            print_stdout(xml, do_force=True)
    else:
        start_install(guest, installer, options)

    return 0


def runcli():  # pragma: no cover
    try:
        sys.exit(main())
    except SystemExit as sys_e:
        sys.exit(sys_e.code)
    except KeyboardInterrupt:
        log.debug("", exc_info=True)
        print_stderr(_("Installation aborted at user request"))
    except Exception as main_e:
        fail(main_e)



#!/usr/bin/env python3

import argparse

parser = argparse.ArgumentParser()
parser.add_argument("sharepath")
parser.add_argument("pkgname")
parser.add_argument("filename")

args = parser.parse_args()

print(
    f"""#!/usr/bin/env python3

import sys
sys.path.insert(0, "{args.sharepath}")
from {args.pkgname} import {args.filename}

{args.filename}.runcli()"""
)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import unittest.mock

import tests
from . import lib


###################
# Private helpers #
###################


def _open_newvm(app):
    button = app.root.find("New", "push button")
    # Launching the dialog can be very flakey without this explicit
    # point() call, not sure why
    button.point()
    button.click()
    return app.find_window("New VM")


def _nav(newvm, forward, back, check):
    pagenumlabel = newvm.find("pagenum-label")
    oldtext = pagenumlabel.text
    ignore = back

    # Clicking is tough to manage, because when clicking
    # rapidly in succession the create wizard has a few
    # cases of stealing focus to better help the user
    # navigate the wizard, but this is very racy and
    # tough to deal with hear. Luckily accelerators
    # don't care too much about focus
    if forward:
        button = newvm.find("Forward", "push button")
        combo = "<alt>f"
    else:
        button = newvm.find("Back", "push button")
        combo = "<alt>b"

    button.check_onscreen()
    button.keyCombo(combo)
    if check:
        lib.utils.check(lambda: pagenumlabel.text != oldtext)


def _forward(newvm, check=True):
    _nav(newvm, forward=True, back=False, check=check)


def _back(newvm, check=True):
    _nav(newvm, forward=False, back=True, check=check)


############################################
# UI tests for virt-manager's NewVM wizard #
############################################


def testNewVMMultiConnection(app):
    """
    Test the wizard's multiple connection handling
    """
    manager = app.topwin

    # Check the dialog shows 'no connection' error
    app.sleep(1)  # give some time for the connection to connect
    manager.grab_focus()
    app.manager_conn_disconnect("test testdriver.xml")
    newvm = _open_newvm(app)
    newvm.find_fuzzy("No active connection to install on")
    newvm.window_close()
    lib.utils.check(lambda: manager.active)

    # Check the xen PV only startup warning
    def _capsopt(fname):
        capsdir = tests.utils.DATADIR + "/capabilities/"
        return ",caps=" + capsdir + fname

    def _add_conn(uri):
        return app.manager_createconn(uri)

    # Test empty qemu connection
    _add_conn(tests.utils.URIs.kvm_x86 + _capsopt("test-empty.xml"))
    newvm = _open_newvm(app)
    newvm.find(".*No hypervisor options were found.*KVM kernel modules.*")
    newvm.window_close()
    app.manager_conn_disconnect("QEMU/KVM")

    _add_conn(tests.utils.URIs.kvm_x86_session + _capsopt("test-qemu-no-kvm.xml"))
    newvm = _open_newvm(app)
    newvm.find(".*KVM is not available.*")
    newvm.window_close()

    _add_conn(tests.utils.URIs.lxc)
    _add_conn(tests.utils.URIs.test_full)
    _add_conn(tests.utils.URIs.test_default)

    # Open the new VM wizard, select a connection
    newvm = _open_newvm(app)
    newvm.combo_select("create-conn", ".*testdriver.xml.*")
    _forward(newvm)

    # Verify media-combo contents for testdriver.xml
    cdrom = newvm.find("media-combo")
    entry = newvm.find("media-entry")
    cdrom.click_combo_entry()
    cdrom.find_fuzzy(r"\(/dev/sr1\)")
    entry.click()
    # Launch this so we can verify storage browser is reset too
    newvm.find_fuzzy("install-iso-browse", "button").click()
    app.select_storagebrowser_volume("pool-dir", "iso-vol")
    newvm.find_fuzzy("Automatically detect", "check").click()
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm)

    # Back up, select test:///default, verify media-combo is now empty
    newvm.window_close()
    newvm = _open_newvm(app)
    newvm.combo_select("create-conn", ".*test default.*")
    _forward(newvm)
    cdrom.click_combo_entry()
    lib.utils.check(lambda: "/dev/sr1" not in cdrom.fmt_nodes())
    app.rawinput.pressKey("Escape")
    newvm.find_fuzzy("install-iso-browse", "button").click()
    browsewin = app.root.find("vmm-storage-browser")
    lib.utils.check(lambda: "pool-logical" not in browsewin.fmt_nodes())


def testNewVMManualDefault(app):
    """
    Click through the New VM wizard with default values + manual, then
    delete the VM
    """
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Manual", "radio").click()
    _forward(newvm)
    osentry = newvm.find("oslist-entry")
    lib.utils.check(lambda: not osentry.text)

    # Make sure we throw an error if no OS selected
    _forward(newvm, check=False)
    app.click_alert_button("You must select", "OK")

    # Test activating the osentry to grab the popover selection
    osentry.click()
    osentry.typeText("generic")
    newvm.find("oslist-popover")
    osentry.click()
    app.rawinput.pressKey("Enter")
    lib.utils.check(lambda: "Generic" in osentry.text)

    # Verify back+forward still keeps Generic selected
    _back(newvm)
    _forward(newvm)
    lib.utils.check(lambda: "Generic" in osentry.text)
    osentry.check_onscreen()
    _forward(newvm)
    _forward(newvm)
    _forward(newvm)

    # Empty triggers a specific codepath
    newvm.find_fuzzy("Name", "text").set_text("")
    # Name collision failure
    newvm.find_fuzzy("Name", "text").set_text("test-many-devices")
    newvm.find_fuzzy("Finish", "button").click()
    app.click_alert_button("in use", "OK")
    newvm.find_fuzzy("Name", "text").set_text("vm1")
    newvm.find_fuzzy("Finish", "button").click()

    # Delete it from the VM window
    vmwindow = app.find_details_window("vm1")
    vmwindow.find("Virtual Machine", "menu").click()
    vmwindow.find("Delete", "menu item").click()

    delete = app.find_window("Delete")
    delete.find_fuzzy("Delete", "button").click()
    app.click_alert_button("Are you sure", "Yes")

    # Verify delete dialog and VM dialog are now gone
    lib.utils.check(lambda: vmwindow.showing is False)


def testNewVMStorage(app):
    """
    Test some storage specific paths
    """
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Manual", "radio").click()
    _forward(newvm)
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm)
    _forward(newvm)

    # qcow2 default shouldn't trigger size error
    sizetext = newvm.find(None, "spin button", "GiB")
    sizetext.set_text("10000000")
    _forward(newvm)
    _back(newvm)

    # Use the storage browser to select a local file
    storagetext = newvm.find("storage-entry")
    newvm.find_fuzzy("Select or create", "radio").click()
    newvm.find("storage-browse").click()
    browse = app.root.find("vmm-storage-browser")
    browse.find("Browse Local", "push button").click()
    chooser = app.root.find("Locate existing storage", "file chooser")
    fname = "COPYING"
    chooser.find(fname, "table cell").click()
    chooser.find("Open", "push button").click()
    lib.utils.check(lambda: newvm.active)
    lib.utils.check(lambda: "COPYING" in storagetext.text)

    # Start the install
    _forward(newvm)
    newvm.find("Finish", "push button").click()
    app.find_details_window("vm1")
    lib.utils.check(lambda: not newvm.showing)


def testNewVMCDROMRegular(app):
    """
    Create a new CDROM VM, choosing distro win8, and do some basic
    'Customize before install' before exiting
    """
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Local install media", "radio").click()
    _forward(newvm)

    # check prepopulated cdrom media
    combo = newvm.find("media-combo")
    combo.click_combo_entry()
    combo.find(r"No media detected \(/dev/sr1\)")
    combo.find(r"Fedora12_media \(/dev/sr0\)").click()

    # Catch validation error
    entry = newvm.find("media-entry")
    lib.utils.check(lambda: "/dev/sr0" in entry.text)
    entry.click()
    entry.set_text("")
    # Something about entry.set_text is flakey with focus,
    # this stuff is to try and fix focus
    app.rawinput.pressKey("Escape")
    newvm.click_title()
    _forward(newvm, check=False)
    app.click_alert_button("media selection is required", "OK")

    # test entry activation too
    entry.click()
    entry.set_text("/dev/sr0")
    app.rawinput.pressKey("Enter")

    # Select a fake iso
    newvm.find_fuzzy("install-iso-browse", "button").click()
    app.select_storagebrowser_volume("pool-dir", "iso-vol")

    osentry = newvm.find("oslist-entry")
    lib.utils.check(lambda: osentry.text == "None detected")

    # Change distro to win8
    newvm.find_fuzzy("Automatically detect", "check").click()
    osentry.click()
    osentry.set_text("windows 8")
    popover = newvm.find("oslist-popover")
    popover.check_onscreen()
    # Verify Escape resets the text entry
    app.rawinput.pressKey("Escape")
    popover.check_not_onscreen()
    lib.utils.check(lambda: osentry.text == "")
    # Re-enter text
    osentry.set_text("windows 8")
    popover.check_onscreen()
    popover.find_fuzzy("include-eol").click()
    popover.find_fuzzy(r"\(win8\)").click()
    popover.check_not_onscreen()
    foundtext = osentry.text
    # Start typing again, and exit, make sure it resets to previous entry
    osentry.click()
    osentry.set_text("foo")
    popover.check_onscreen()
    app.rawinput.pressKey("Escape")
    popover.check_not_onscreen()
    lib.utils.check(lambda: osentry.text == foundtext)
    _forward(newvm)

    # Verify that CPU values are non-default
    cpus = newvm.find("cpus", "spin button")
    lib.utils.check(lambda: int(cpus.text) > 1, timeout=5)
    _forward(newvm)
    _forward(newvm)

    # Select customize wizard
    newvm.find_fuzzy("Customize", "check").click()
    newvm.find_fuzzy("Finish", "button").click()

    # Verify CDROM media is inserted
    vmwindow = app.find_details_window("win8")
    vmwindow.find_fuzzy("IDE CDROM", "table cell").click()
    mediaent = vmwindow.find("media-entry")
    lib.utils.check(lambda: "iso-vol" in mediaent.text)

    # Change boot autostart
    vmwindow.find_fuzzy("Boot", "table cell").click()
    vmwindow.find_fuzzy("Start virtual machine", "check").click()
    vmwindow.find_fuzzy("config-apply").click()

    # Change to 'copy host CPU'
    vmwindow.find_fuzzy("CPUs", "table cell").click()
    cpucheck = vmwindow.find_fuzzy("Copy host", "check")
    cpucheck.click()
    vmwindow.find_fuzzy("config-apply").click()
    lib.utils.check(lambda: "host-passthrough" in cpucheck.name)

    # Add a default disk
    vmwindow.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    addhw.find("Finish", "push button").click()
    lib.utils.check(lambda: vmwindow.active)

    # Select the new disk, change the bus to USB
    vmwindow.find_fuzzy("IDE Disk 2", "table cell").click()
    appl = vmwindow.find("config-apply", "push button")
    hwlist = vmwindow.find("hw-list")
    tab = vmwindow.find("disk-tab")
    tab.find("Disk bus:", "text").set_text("usb")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    # Device is now 'USB Disk 1'
    c = hwlist.find("USB Disk 1", "table cell")
    lib.utils.check(lambda: c.state_selected)
    tab.find("Advanced options", "toggle button").click_expander()
    tab.find("Removable:", "check box").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Change NIC mac
    vmwindow.find_fuzzy("NIC", "table cell").click()
    tab = vmwindow.find("network-tab")
    tab.find("mac-entry", "text").set_text("00:11:00:11:00:11")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Start the install, close via the VM window
    vmwindow.find_fuzzy("Begin Installation", "button").click()
    lib.utils.check(lambda: newvm.showing is False)
    vmwindow = app.find_details_window("win8")
    vmwindow.find_fuzzy("File", "menu").click()
    vmwindow.find_fuzzy("Quit", "menu item").click()
    lib.utils.check(lambda: app.is_running())


def testNewVMCDROMDetect(app):
    """
    CDROM with detection
    """
    cdrom = tests.utils.DATADIR + "/fakemedia/fake-win7.iso"
    newvm = _open_newvm(app)
    newvm.find_fuzzy("Local install media", "radio").click()
    _forward(newvm)
    newvm.find("media-entry").click()
    newvm.find("media-entry").set_text(cdrom)
    # Use forward to trigger detection
    _forward(newvm)
    _forward(newvm)
    _forward(newvm)
    newvm.find("Finish", "push button").click()
    app.find_details_window("win7")
    lib.utils.check(lambda: not newvm.showing)


def testNewVMURL(app):
    """
    New VM with URL and distro detection, plus having fun with
    the storage browser and network selection.
    """
    # Also test default UEFI from prefs
    app.open(keyfile="uefi.ini", uri=tests.utils.URIs.kvm_x86)
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Network Install", "radio").click()
    _forward(newvm)
    osentry = newvm.find("oslist-entry")
    lib.utils.check(lambda: osentry.text.startswith("Waiting"))

    newvm.find("install-url-entry").set_text("")
    _forward(newvm, check=False)
    app.click_alert_button("tree is required", "OK")

    url = (
        "https://archives.fedoraproject.org/pub/archive/fedora/linux/releases/10/Fedora/x86_64/os/"
    )
    oslabel = "Fedora 10"
    newvm.find("install-url-entry").set_text(url)
    newvm.find("install-url-entry").click()
    app.rawinput.pressKey("Enter")
    newvm.find("install-urlopts-expander").click_expander()
    newvm.find("install-urlopts-entry").set_text("foo=bar")

    lib.utils.check(lambda: osentry.text == oslabel, timeout=10)

    # Move forward, then back, ensure OS stays selected
    _forward(newvm)
    _back(newvm)
    lib.utils.check(lambda: osentry.text == oslabel)

    # Disable autodetect, make sure OS still selected
    newvm.find_fuzzy("Automatically detect", "check").click()
    lib.utils.check(lambda: osentry.text == oslabel)
    _forward(newvm)
    _back(newvm)

    # Ensure the EOL field was selected
    osentry.click()
    app.rawinput.pressKey("Down")
    popover = newvm.find("oslist-popover")
    lib.utils.check(lambda: popover.showing)
    includeeol = newvm.find("include-eol", "check")
    lib.utils.check(lambda: includeeol.isChecked)

    # Re-enable autodetect, check for detecting text
    newvm.find_fuzzy("Automatically detect", "check").click()
    lib.utils.check(lambda: not popover.showing)
    lib.utils.check(lambda: "Detecting" in osentry.text)
    lib.utils.check(lambda: osentry.text == oslabel, timeout=10)

    # Progress the install
    _forward(newvm)
    _forward(newvm)
    _forward(newvm)
    newvm.find_fuzzy("Finish", "button").click()

    progress = app.find_window("Creating Virtual Machine")
    lib.utils.check(lambda: not progress.showing, timeout=120)

    details = app.find_details_window("fedora10")
    lib.utils.check(lambda: not newvm.showing)

    # Re-run the newvm wizard, check that URL was remembered
    details.window_close()
    newvm = _open_newvm(app)
    newvm.find_fuzzy("Network Install", "radio").click()
    _forward(newvm)
    urlcombo = newvm.find("install-url-combo")
    lib.utils.check(lambda: urlcombo.showing)
    lib.utils.check(lambda: url in urlcombo.fmt_nodes())


def testNewKVMQ35Tweaks(app):
    """
    New VM that should default to Q35, but tweak things a bunch
    """
    app.uri = tests.utils.URIs.kvm_x86
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Import", "radio").click()
    _forward(newvm)
    newvm.find("import-entry").set_text("/pool-dir/testvol1.img")
    newvm.find("oslist-entry").set_text("fribfrob")
    popover = newvm.find("oslist-popover")
    osname = "linux2022"
    popover.find_fuzzy(osname).click()
    _forward(newvm)
    _forward(newvm)

    # Select customize wizard, we will use this VM to
    # hit some code paths elsewhere
    newvm.find_fuzzy("Customize", "check").click()
    newvm.find_fuzzy("Finish", "button").click()
    vmname = osname
    details = app.find_details_window(vmname)
    appl = details.find("config-apply")

    # Tweak some Overview settings
    details.combo_check_default("Chipset:", "Q35")
    details.combo_check_default("Firmware:", "BIOS")

    # Unchanged machine
    details.combo_select("Chipset:", "i440FX")
    details.combo_select("Chipset:", "Q35")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    # Switch i440FX
    details.combo_select("Chipset:", "i440FX")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    # Switch back to Q35
    details.combo_select("Chipset:", "Q35")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    # Switch to UEFI, back to BIOS, back to UEFI
    details.combo_select("Firmware:", ".*x86_64.*")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    # Switch back to BIOS
    details.combo_select("Firmware:", "BIOS")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    # Switch back to UEFI
    details.combo_select("Firmware:", ".*x86_64.*")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Verify host-passthrough selected
    details.find_fuzzy("CPUs", "table cell").click()
    cpucheck = details.find_fuzzy("Copy host", "check")
    assert "host-passthrough" in cpucheck.name
    new_xml = lib.utils.get_xmleditor_xml(app, details)
    assert "host-passthrough" in new_xml

    # Add another network device
    details.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    addhw.find("Network", "table cell").click()
    tab = addhw.find("network-tab", None)
    lib.utils.check(lambda: tab.showing)
    addhw.find("Finish", "push button").click()
    lib.utils.check(lambda: not addhw.active)
    lib.utils.check(lambda: details.active)

    # Finish
    details.find_fuzzy("Begin Installation", "button").click()
    lib.utils.check(lambda: details.dead)
    app.find_details_window(vmname)


def testNewKVMQ35UEFI(app):
    """
    New VM that should default to Q35, and set UEFI
    """
    app.uri = tests.utils.URIs.kvm_x86
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Import", "radio").click()
    _forward(newvm)
    newvm.find("import-entry").set_text("/pool-dir/testvol1.img")
    newvm.find("oslist-entry").set_text("fedora30")
    popover = newvm.find("oslist-popover")
    popover.find("include-eol").click()
    popover.find_fuzzy("Fedora 30").click()
    _forward(newvm)
    _forward(newvm)

    # Select customize wizard, we will use this VM to
    # hit some PPC64 code paths elsewhere
    newvm.find_fuzzy("Customize", "check").click()
    newvm.find_fuzzy("Finish", "button").click()
    vmname = "fedora30"
    details = app.find_details_window(vmname)

    # Change to UEFI
    details.combo_check_default("Chipset:", "Q35")
    details.combo_check_default("Firmware:", "BIOS")
    details.combo_select("Firmware:", "UEFI")
    details.find("config-apply").click()
    new_xml = lib.utils.get_xmleditor_xml(app, details)
    assert 'os firmware="efi"' in new_xml

    # Finish
    details.find_fuzzy("Begin Installation", "button").click()
    lib.utils.check(lambda: details.dead)
    app.find_details_window(vmname)


def testNewPPC64(app):
    """
    New PPC64 VM to test architecture selection
    """
    app.uri = tests.utils.URIs.kvm_x86
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Architecture options", "toggle").click()
    newvm.combo_select("Architecture", ".*ppc64.*")
    newvm.combo_check_default("Machine Type", ".*pseries.*")

    newvm.find_fuzzy("Manual", "radio").click()
    _forward(newvm)
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm)
    _forward(newvm)
    # Disable storage, we add some via customize
    newvm.find_fuzzy("Enable storage", "check box").click()
    _forward(newvm)

    # Select customize wizard, we will use this VM to
    # hit some PPC64 code paths elsewhere
    newvm.find_fuzzy("Customize", "check").click()
    newvm.find_fuzzy("Finish", "button").click()
    details = app.find_details_window("vm-ppc64")

    tab = details.find("overview-tab")
    tab.combo_check_default("machine-combo", "pseries")
    tab.combo_select("machine-combo", "pseries-2.1")
    appl = details.find("config-apply")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Add a TPM SPAPR device
    details.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    addhw.find("TPM", "table cell").click()
    tab = addhw.find("tpm-tab", None)
    lib.utils.check(lambda: tab.showing)
    addhw.find("Finish", "push button").click()
    lib.utils.check(lambda: not addhw.active)
    lib.utils.check(lambda: details.active)

    # Add a SCSI disk which also adds virtio-scsi controller
    details.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    addhw.find("Storage", "table cell").click()
    tab = addhw.find("storage-tab", None)
    lib.utils.check(lambda: tab.showing)
    tab.combo_select("Bus type:", "SCSI")
    addhw.find("Finish", "push button").click()
    lib.utils.check(lambda: not addhw.active)
    lib.utils.check(lambda: details.active)

    # Finish
    details.find_fuzzy("Begin Installation", "button").click()
    lib.utils.check(lambda: details.dead)
    app.find_details_window("vm-ppc64")


def testNewVMAArch64UEFI(app):
    """
    Test aarch64 UEFI usage
    """
    app.uri = tests.utils.URIs.kvm_aarch64
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Local install media", "radio").click()
    _forward(newvm)

    newvm.find_fuzzy("Automatically detect", "check").click()
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    newvm.find("media-entry").set_text("/pool-dir/testvol1.img")
    _forward(newvm)
    _forward(newvm)
    # Disable storage, this triggers a livecd code path in createvm.py
    newvm.find_fuzzy("Enable storage", "check box").click()
    _forward(newvm)
    newvm.find_fuzzy("Finish", "button").click()

    app.find_details_window("vm1")
    lib.utils.check(lambda: not newvm.showing)


def testNewVMArmKernel(app):
    """
    New arm VM that requires kernel/initrd/dtb
    """
    app.uri = tests.utils.URIs.kvm_armv7l_nodomcaps
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Architecture options", "toggle").click_expander()
    newvm.find_fuzzy("Virt Type", "combo").click()
    KVM = newvm.find_fuzzy("KVM", "menu item")
    TCG = newvm.find_fuzzy("TCG", "menu item")
    lib.utils.check(lambda: KVM.selected)
    lib.utils.check(lambda: TCG.showing)
    app.rawinput.pressKey("Esc")

    # Validate some initial defaults
    local = newvm.find_fuzzy("Local", "radio")
    lib.utils.check(lambda: not local.sensitive)
    newvm.find_fuzzy("Machine Type", "combo").click()
    newvm.find_fuzzy("canon", "menu item").click()
    newvm.find_fuzzy("Machine Type", "combo").click()
    newvm.find("virt", "menu item").click()
    importradio = newvm.find("Import", "radio")
    importradio.click()
    lib.utils.check(lambda: importradio.checked)
    _forward(newvm)

    newvm.find("import-entry").set_text("/pool-dir/default-vol")
    # Make sure the info box shows up
    newvm.find("Kernel/initrd settings can be configured")
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm, check=False)

    # Disk collision box pops up, hit ok
    app.click_alert_button("in use", "Yes")

    _forward(newvm)
    newvm.find_fuzzy("Finish", "button").click()

    lib.utils.check(lambda: not newvm.showing)
    app.find_details_window("vm1")


def testNewVMContainerApp(app):
    """
    Simple LXC app install
    """
    app.uri = tests.utils.URIs.lxc

    newvm = _open_newvm(app)
    newvm.find_fuzzy("Application", "radio").click()
    _forward(newvm)

    # Set custom init
    apptext = newvm.find_fuzzy(None, "text", "application path")
    apptext.set_text("")
    _forward(newvm, check=False)
    app.click_alert_button("path is required", "OK")
    newvm.find("install-app-browse").click()
    app.select_storagebrowser_volume("pool-dir", "aaa-unused.qcow2")
    lib.utils.check(lambda: "aaa-unused.qcow2" in apptext.text)

    _forward(newvm)
    _forward(newvm)
    # Trigger back, to ensure disk page skipping works
    _back(newvm)
    _back(newvm)
    _forward(newvm)
    _forward(newvm)

    # Select customize wizard, we will use this VM to hit specific
    # code paths
    newvm.find_fuzzy("Customize", "check").click()
    newvm.find_fuzzy("Finish", "button").click()
    vmname = "container1"
    details = app.find_details_window(vmname)

    # Tweak init values
    details.find("Boot Options", "table cell").click()
    tab = details.find("boot-tab")
    tab.find("Init path:", "text").set_text("")
    tab.find("Init args:", "text").set_text("some args")
    appl = details.find("config-apply")
    appl.click()
    app.click_alert_button("init path must be specified", "OK")
    lib.utils.check(lambda: appl.sensitive)
    tab.find("Init path:", "text").set_text("/some/path")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Check that addhw container options are disabled
    details.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    addhw.find("PCI Host Device", "table cell").click()
    # Ensure the error label is showing
    label = addhw.find("Not supported for containers")
    label.check_onscreen()
    addhw.find("Cancel", "push button").click()
    lib.utils.check(lambda: not addhw.active)
    lib.utils.check(lambda: details.active)

    # Finish
    details.find_fuzzy("Begin Installation", "button").click()
    lib.utils.check(lambda: not newvm.showing)
    app.find_details_window(vmname)


def testNewVMCustomizeCancel(app):
    """
    Test cancelling out of the customize wizard
    """
    newvm = _open_newvm(app)
    newvm.find_fuzzy("Manual", "radio").click()
    _forward(newvm)
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm)
    _forward(newvm)
    _forward(newvm)

    newvm.find_fuzzy("Customize", "check").click()
    newvm.find_fuzzy("Finish", "button").click()
    vmname = "vm1"
    details = app.find_details_window(vmname)

    details.find("Cancel Installation", "push button").click()
    app.click_alert_button("abort the installation", "No")
    lib.utils.check(lambda: details.active)
    details.find("Cancel Installation", "push button").click()
    app.click_alert_button("abort the installation", "Yes")
    lib.utils.check(lambda: not details.active)
    lib.utils.check(lambda: not newvm.active)


def testNewVMCustomizeMisc(app):
    """
    Some specific customize logic paths
    """
    app.open(keyfile="rawdefault.ini")
    newvm = _open_newvm(app)
    newvm.find_fuzzy("Manual", "radio").click()
    _forward(newvm)
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm)
    _forward(newvm)

    # Raw default will be non-sparse, should trigger size error
    sizetext = newvm.find(None, "spin button", "GiB")
    sizetext.set_text("10000000")
    _forward(newvm, check=False)
    app.click_alert_button("Storage parameter error", "OK")
    sizetext.set_text("1")

    _forward(newvm)

    newvm.find_fuzzy("Customize", "check").click()
    newvm.find_fuzzy("Finish", "button").click()
    vmname = "vm1"
    details = app.find_details_window(vmname)

    # Test name change
    tab = details.find("overview-tab")
    nametext = tab.find("Name:", "text")
    nametext.set_text("foonewname")
    details.find("config-apply").click()
    app.find_details_window("foonewname")

    # Trigger XML failure to hit some codepaths
    nametext.set_text("")
    details.find("Begin Installation").click()
    app.click_alert_button("unapplied changes", "Yes")
    app.click_alert_button("name must be specified", "Close")
    lib.utils.check(lambda: details.showing)

    # Discard XML change and continue with install
    details.find("Begin Installation").click()
    app.click_alert_button("unapplied changes", "No")
    lib.utils.check(lambda: not details.showing)
    lib.utils.check(lambda: not newvm.showing)
    app.find_details_window("foonewname")


def testNewVMContainerTree(app):
    """
    Simple LXC tree install
    """
    app.uri = tests.utils.URIs.lxc

    newvm = _open_newvm(app)
    newvm.find_fuzzy("Operating system", "radio").click()
    _forward(newvm)

    # Set directory path
    dirtext = newvm.find_fuzzy(None, "text", "root directory")
    dirtext.set_text("")
    _forward(newvm, check=False)
    app.click_alert_button("path is required", "OK")

    newvm.find("install-oscontainer-browse").click()
    app.select_storagebrowser_volume("pool-dir", "dir-vol")
    lib.utils.check(lambda: "dir-vol" in dirtext.text)

    _forward(newvm)
    _forward(newvm)
    newvm.find_fuzzy("Finish", "button").click()

    lib.utils.check(lambda: not newvm.showing)
    app.find_details_window("container1")


def testNewVMContainerVZ(app):
    """
    Virtuozzo container install
    """
    app.uri = tests.utils.URIs.vz

    newvm = _open_newvm(app)
    newvm.find_fuzzy("Container", "radio").click()
    newvm.find_fuzzy("Virtual machine", "radio").click()
    newvm.find_fuzzy("Container", "radio").click()
    _forward(newvm)

    # Set directory path
    templatetext = newvm.find_fuzzy(None, "text", "container template")
    templatetext.set_text("")
    _forward(newvm, check=False)
    app.click_alert_button("template name is required", "OK")
    templatetext.set_text("centos-6-x86_64")
    _forward(newvm)
    _forward(newvm)
    newvm.find_fuzzy("Finish", "button").click()

    app.find_details_window("container1")
    lib.utils.check(lambda: not newvm.showing)


def testNewVMContainerBootstrap(app):
    app.open(uri=tests.utils.URIs.lxc, extra_opts=["--test-options=fake-virtbootstrap"])

    newvm = _open_newvm(app)
    newvm.find_fuzzy("Operating system", "radio").click()
    _forward(newvm)

    # Set directory path
    import tempfile

    tmpdir = tempfile.TemporaryDirectory()
    newvm.find_fuzzy("Create OS directory", "check box").click()

    uritext = newvm.find("install-oscontainer-source-uri")
    uritext.text = ""
    _forward(newvm, check=False)
    app.click_alert_button("Source URL is required", "OK")
    uritext.text = "docker://alpine"

    rootdir = newvm.find_fuzzy(None, "text", "root directory")
    lib.utils.check(lambda: ".local/share/libvirt" in rootdir.text)
    rootdir.set_text("/dev/null")
    _forward(newvm, check=False)
    app.click_alert_button("not directory", "OK")
    rootdir.set_text("/root")
    _forward(newvm, check=False)
    app.click_alert_button("No write permissions", "OK")
    rootdir.set_text("/tmp")
    _forward(newvm, check=False)
    app.click_alert_button("directory is not empty", "No")
    rootdir.set_text(tmpdir.name)
    newvm.find("install-oscontainer-root-passwd").set_text("foobar")
    # Invalid credentials to trigger failure
    newvm.find("Credentials", "toggle button").click_expander()
    newvm.find("bootstrap-registry-user").set_text("foo")
    _forward(newvm, check=None)
    app.click_alert_button("Please specify password", "OK")
    newvm.find("bootstrap-registry-password").set_text("bar")

    _forward(newvm)
    _forward(newvm)
    newvm.find_fuzzy("Finish", "button").click()
    app.click_alert_button("virt-bootstrap did not complete", "Close")
    _back(newvm)
    _back(newvm)
    newvm.find("bootstrap-registry-user").set_text("")
    newvm.find("bootstrap-registry-password").set_text("")

    _forward(newvm)
    _forward(newvm)
    newvm.find_fuzzy("Finish", "button").click()
    prog = app.find_window("Creating Virtual Machine")
    lib.utils.check(lambda: not prog.showing, timeout=30)

    lib.utils.check(lambda: not newvm.showing)
    app.find_details_window("container1")


def testNewVMXenPV(app):
    """
    Test the create wizard with a fake xen PV install
    """
    app.uri = tests.utils.URIs.xen
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Architecture options", "toggle").click()
    newvm.combo_select("Xen Type", ".*paravirt.*")

    newvm.find_fuzzy("Import", "radio").click()
    _forward(newvm)
    newvm.find("import-entry").set_text("/pool-dir/testvol1.img")
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm)
    _forward(newvm)
    newvm.find_fuzzy("Finish", "button").click()

    app.find_details_window("vm1")
    lib.utils.check(lambda: not newvm.showing)


def testNewVMInstallFail(app):
    def dofail():
        _newvm = _open_newvm(app)
        _newvm.find_fuzzy("Manual", "radio").click()
        _forward(_newvm)
        _newvm.find("oslist-entry").set_text("generic")
        _newvm.find("oslist-popover").find_fuzzy("generic").click()
        _forward(_newvm)
        _forward(_newvm)
        _forward(_newvm)

        # '/' in name will trigger libvirt error
        _newvm.find_fuzzy("Name", "text").set_text("test/bad")
        _newvm.find_fuzzy("Finish", "button").click()
        app.click_alert_button("Unable to complete install", "Close")
        return _newvm

    newvm = dofail()
    pathlabel = newvm.find(".*test/bad.qcow2")
    generatedpath = pathlabel.text
    # Changing VM name should not generate a new path
    newvm.find_fuzzy("Name", "text").set_text("test/badfoo")
    lib.utils.check(lambda: pathlabel.text == generatedpath)
    newvm.find_fuzzy("Finish", "button").click()
    app.click_alert_button("Unable to complete install", "Close")
    # Closing dialog should trigger storage cleanup path
    newvm.find_fuzzy("Cancel", "button").click()
    lib.utils.check(lambda: not newvm.visible)

    # Run again
    newvm = dofail()
    _back(newvm)
    newvm.find_fuzzy("Select or create", "radio").click()

    newvm.find("storage-entry").set_text("/pool-dir/somenewvol1")
    _forward(newvm)
    newvm.find_fuzzy("Name", "text").set_text("test-foo")
    newvm.find_fuzzy("Finish", "button").click()

    app.find_details_window("test-foo")
    lib.utils.check(lambda: not newvm.showing)


def testNewVMCustomizeXMLEdit(app):
    """
    Test new VM with raw XML editing via customize wizard
    """
    app.open(xmleditor_enabled=True)
    newvm = _open_newvm(app)

    # Create a custom named VM, using CDROM media, and default storage
    vmname = "fooxmleditvm"
    newvm.find_fuzzy("Local install media", "radio").click()
    newvm.find_fuzzy("Forward", "button").click()
    nonexistpath = "/dev/foovmm-idontexist"
    existpath = "/pool-dir/testvol1.img"
    newvm.find("media-entry").set_text(nonexistpath)
    lib.utils.check(lambda: newvm.find("oslist-entry").text == "None detected")
    newvm.find_fuzzy("Automatically detect", "check").click()
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm, check=False)
    app.click_alert_button("Error setting installer", "OK")
    newvm.find("media-entry").set_text(existpath)
    _forward(newvm)
    _forward(newvm)
    _forward(newvm)
    newvm.find_fuzzy("Customize", "check").click()
    newvm.find_fuzzy("Name", "text").set_text(vmname)
    newvm.find_fuzzy("Finish", "button").click()
    win = app.find_details_window(vmname)
    xmleditor = win.find("XML editor")
    finish = win.find("config-apply")

    # Change a device setting with the XML editor
    win.find_fuzzy("IDE Disk 1", "table cell").click()
    tab = win.find("disk-tab")
    win.find("XML", "page tab").click()
    # Change the disk path via the XML editor
    fname = vmname + ".qcow2"
    lib.utils.check(lambda: fname in xmleditor.text)
    newx = xmleditor.text.replace(fname, "default-vol")
    xmleditor.set_text(newx)
    appl = win.find("config-apply")
    # This is kindof a bug, changing path in XML editor in Customize
    # doesn't take effect for storage with creation parameters, but
    # it's a pain to fix.
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    lib.utils.check(lambda: vmname in xmleditor.text)

    # Change a VM setting and verify it
    win.find_fuzzy("Boot", "table cell").click()
    tab = win.find("boot-tab")
    bootmenu = tab.find("Enable boot menu", "check box")
    lib.utils.check(lambda: not bootmenu.checked)
    win.find("XML", "page tab").click()
    newtext = xmleditor.text.replace("<os>", "<os><bootmenu enable='yes'/>")
    xmleditor.set_text(newtext)
    finish.click()
    win.find("Details", "page tab").click()
    lib.utils.check(lambda: bootmenu.checked)

    # Change a device setting with the XML editor
    win.find_fuzzy("NIC", "table cell").click()
    tab = win.find("network-tab")
    win.find("XML", "page tab").click()
    newbrname = "BRFAKE"
    newx = xmleditor.text.replace("network", "bridge")
    newx = newx.replace('bridge="default"', "bridge='%s'" % newbrname)
    xmleditor.set_text(newx)
    finish.click()

    # Finish install.
    win.find_fuzzy("Begin Installation", "button").click()
    lib.utils.check(lambda: win.dead)
    win = app.find_details_window(vmname)
    win.find("Details", "radio button").click()

    # Verify VM change stuck
    win.find_fuzzy("Boot", "table cell").click()
    tab = win.find("boot-tab")
    bootmenu = tab.find("Enable boot menu", "check box")
    lib.utils.check(lambda: bootmenu.checked)

    # Verify device change stuck
    win.find_fuzzy("NIC", "table cell").click()
    tab = win.find("network-tab")
    devname = tab.find("Device name:", "text")
    lib.utils.check(lambda: devname.text == newbrname)

    # Verify install media is handled correctly after XML customize
    win.find_fuzzy("IDE CDROM 1", "table cell").click()
    tab = win.find("disk-tab")
    mediaent = tab.find("media-entry")
    lib.utils.check(lambda: mediaent.text == existpath)
    win.find("Shut Down", "push button").click()
    run = win.find("Run", "push button")
    lib.utils.check(lambda: run.sensitive)
    lib.utils.check(lambda: mediaent.text == "")

    # Verify default disk storage was actually created. This has some
    # special handling in domain.py
    tab.find("Browse", "push button").click()
    browser = app.root.find("vmm-storage-browser")
    browser.find("%s.qcow2" % vmname, "table cell")


def testNewVMRemote(app):
    """
    Hit some is_remote code paths
    """
    app.uri = tests.utils.URIs.test_remote
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Import", "radio").click()
    _forward(newvm)
    importtext = newvm.find("import-entry")

    # Click forward, hitting missing Import path error
    _forward(newvm, check=False)
    app.click_alert_button("import is required", "OK")

    # Click forward, but Import path doesn't exist
    importtext.set_text("/pool-dir/idontexist")
    _forward(newvm, check=False)
    app.click_alert_button("import path must point", "OK")
    importtext.set_text("/pool-dir/default-vol")

    # Click forward, hitting missing OS error
    _forward(newvm, check=False)
    app.click_alert_button("select an OS", "OK")

    # Set OS
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()

    # Click forward, but Import path is in use, and exit
    _forward(newvm, check=False)
    app.click_alert_button("in use", "No")

    # storagebrowser bits
    newvm.find("install-import-browse").click()
    browsewin = app.root.find("vmm-storage-browser")
    # Insensitive for remote connection
    browselocal = browsewin.find("Browse Local")
    lib.utils.check(lambda: browselocal.sensitive is False)
    # Close the browser and reopen
    browsewin.find("Cancel").click()
    lib.utils.check(lambda: not browsewin.active)
    # Reopen, select storage
    newvm.find("install-import-browse").click()
    app.select_storagebrowser_volume("pool-dir", "bochs-vol")
    lib.utils.check(lambda: importtext.text == "/pool-dir/bochs-vol")

    _forward(newvm)
    _forward(newvm)

    newvm.find_fuzzy("Finish", "button").click()
    app.find_details_window("vm1")
    lib.utils.check(lambda: not newvm.showing)


def testNewVMSession(app):
    """
    Test with fake qemu session
    """
    app.uri = tests.utils.URIs.test_empty + ",fakeuri=qemu:///session"
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Import", "radio").click()
    _forward(newvm)
    newvm.find("import-entry").set_text("/etc/hostname")
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm)
    _forward(newvm)
    newvm.combo_check_default("net-source", "Usermode")

    newvm.find_fuzzy("Finish", "button").click()
    details = app.find_details_window("vm1")
    lib.utils.check(lambda: not newvm.showing)
    details.window_close()

    # Ensure disconnecting will close the dialog
    newvm = _open_newvm(app)
    app.manager_test_conn_window_cleanup(".*session.*", newvm)


def testNewVMEmptyConn(app):
    """
    Test with an empty connection
    """
    app.uri = tests.utils.URIs.test_empty
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Import", "radio").click()
    _forward(newvm)
    newvm.find("import-entry").set_text(__file__)
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm)
    _forward(newvm)
    newvm.combo_check_default("net-source", "Bridge")
    warnlabel = newvm.find_fuzzy("suitable default network", "label")
    warnlabel.check_onscreen()
    newvm.find("Device name:", "text").set_text("foobr0")

    # Select customize wizard, we will use this VM to hit specific
    # code paths
    newvm.find_fuzzy("Customize", "check").click()
    newvm.find_fuzzy("Finish", "button").click()
    vmname = "vm1"
    details = app.find_details_window(vmname)

    # Check that addhw hostdev drop down is empty
    details.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    addhw.find("USB Host Device", "table cell").click()
    tab = addhw.find("host-tab", None)
    lib.utils.check(lambda: tab.showing)
    cell = tab.find("No Devices", "table cell")
    lib.utils.check(lambda: cell.selected)
    addhw.find("Cancel", "push button").click()
    lib.utils.check(lambda: not addhw.active)
    lib.utils.check(lambda: details.active)

    # Finish
    details.find_fuzzy("Begin Installation", "button").click()
    lib.utils.check(lambda: details.dead)
    app.find_details_window(vmname)


def testNewVMInactiveNetwork(app):
    """
    Test with an inactive 'default' network
    """
    app.uri = tests.utils.URIs.test_default
    hostwin = app.manager_open_host("Virtual Networks", conn_label="test default")
    cell = hostwin.find("default", "table cell")
    cell.click()
    hostwin.find("net-stop").click()
    hostwin.keyCombo("<ctrl>w")

    newvm = _open_newvm(app)

    newvm.find_fuzzy("Import", "radio").click()
    _forward(newvm)
    newvm.find("import-entry").set_text(__file__)
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm)
    _forward(newvm)

    newvm.find_fuzzy("Finish", "button").click()
    app.click_alert_button("start the network", "Yes")
    lib.utils.check(lambda: not newvm.showing)


@unittest.mock.patch.dict("os.environ", {"VIRTINST_TEST_SUITE": "1"})
def testNewVMDefaultBridge(app):
    """
    We actually set the unittest env variable here, which
    sets a fake bridge in interface.py
    """
    app.uri = tests.utils.URIs.test_empty
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Import", "radio").click()
    _forward(newvm)
    newvm.find("import-entry").set_text(__file__)
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm)
    _forward(newvm)
    newvm.find("Network selection", "toggle button").click_expander()
    newvm.combo_check_default("net-source", "Bridge")
    devname = newvm.find("Device name:", "text")
    lib.utils.check(lambda: devname.text == "testsuitebr0")

    newvm.find_fuzzy("Finish", "button").click()
    app.find_details_window("vm1")
    lib.utils.check(lambda: not newvm.showing)


@unittest.mock.patch.dict("os.environ", {"VIRTINST_TEST_SUITE_FAKE_NO_SPICE": "1"})
def testCreateVMMissingSpice(app):
    newvm = _open_newvm(app)

    newvm.find_fuzzy("Import", "radio").click()
    _forward(newvm)
    newvm.find("import-entry").set_text("/pool-dir/testvol1.img")
    newvm.find("oslist-entry").set_text("generic")
    newvm.find("oslist-popover").find_fuzzy("generic").click()
    _forward(newvm)
    _forward(newvm)

    newvm.find_fuzzy("Finish", "button").click()
    details = app.find_details_window("vm1")
    lib.utils.check(lambda: not newvm.showing)

    details.find_fuzzy("test suite faking no spice")
    details.window_close()



# Copyright (C) 2015 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import pytest

from virtinst import URI

import tests


############################
# Test virtinst URI module #
############################


def _compare(
    uri,
    scheme="",
    transport="",
    port="",
    username="",
    path="",
    hostname="",
    query="",
    fragment="",
    is_ipv6=False,
    host_is_ipv4_string=False,
):
    uriinfo = URI(uri)
    assert scheme == uriinfo.scheme
    assert transport == uriinfo.transport
    assert port == uriinfo.port
    assert username == uriinfo.username
    assert path == uriinfo.path
    assert hostname == uriinfo.hostname
    assert query == uriinfo.query
    assert fragment == uriinfo.fragment
    assert is_ipv6 == uriinfo.is_ipv6
    assert host_is_ipv4_string == uriinfo.host_is_ipv4_string


def testURIs():
    _compare("lxc://", scheme="lxc")
    _compare("qemu:///session", scheme="qemu", path="/session")
    _compare(
        "http://foobar.com:5901/my/example.path#my-frag",
        scheme="http",
        hostname="foobar.com",
        port="5901",
        path="/my/example.path",
        fragment="my-frag",
    )
    _compare(
        "gluster+tcp://[1:2:3:4:5:6:7:8]:24007/testvol/dir/a.img",
        scheme="gluster",
        transport="tcp",
        hostname="1:2:3:4:5:6:7:8",
        port="24007",
        path="/testvol/dir/a.img",
        is_ipv6=True,
    )
    _compare(
        "qemu+ssh://root@192.168.2.3/system?no_verify=1",
        scheme="qemu",
        transport="ssh",
        username="root",
        hostname="192.168.2.3",
        path="/system",
        query="no_verify=1",
        host_is_ipv4_string=True,
    )
    _compare(
        "qemu+ssh://foo%5Cbar@hostname/system",
        scheme="qemu",
        path="/system",
        transport="ssh",
        hostname="hostname",
        username="foo\\bar",
    )
    _compare(
        "qemu+ssh://user%40domain.org@hostname/system",
        scheme="qemu",
        path="/system",
        transport="ssh",
        hostname="hostname",
        username="user@domain.org",
    )


def test_magicuri_connver():
    uri = tests.utils.URIs.test_default + ",connver=1,libver=2"
    conn = tests.utils.URIs.openconn(uri)
    assert conn.conn_version() == 1
    assert conn.local_libvirt_version() == 2

    conn = tests.utils.URIs.openconn("test:///default")
    # Add some support tests with it
    with pytest.raises(ValueError, match=".*type <class 'libvirt.virDomain'>.*"):
        conn.support.domain_xml_inactive("foo")

    # pylint: disable=protected-access
    from virtinst import support

    def _run(**kwargs):
        check = support._SupportCheck(**kwargs)
        return check(conn)

    assert _run(function="virNope.Foo") is False
    assert _run(function="virDomain.IDontExist") is False
    assert _run(function="virDomain.isActive") is True
    assert _run(function="virConnect.getVersion", flag="SOME_FLAG_DOESNT_EXIST") is False
    assert _run(version="1000.0.0") is False
    assert _run(hv_version={"test": "1000.0.0"}) is False
    assert _run(hv_libvirt_version={"test": "1000.0.0"}) is False
    assert _run(hv_libvirt_version={"qemu": "1.2.3"}) is False
    assert _run(hv_libvirt_version={"qemu": "1.2.3", "all": 0}) is True

    dom = conn.lookupByName("test")
    assert conn.support.domain_xml_inactive(dom) is True
